#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for control volumes
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# We use some private attributes here to hide these from the user
# pylint: disable=protected-access

# Import Python libraries
import copy
from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    Expression,
    Param,
    Reals,
    TransformationFactory,
    units as pyunits,
    Var,
    Reference,
    value,
)
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.common.config import ConfigValue, In
from pyomo.common.deprecation import deprecation_warning

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    ControlVolumeBlockData,
    FlowDirection,
    MaterialFlowBasis,
    MaterialBalanceType,
)
from idaes.core.util.exceptions import (
    BalanceTypeNotSupportedError,
    ConfigurationError,
    PropertyNotSupportedError,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.config import is_transformation_method, is_transformation_scheme
from idaes.core.util import scaling as iscale

import idaes.logger as idaeslog

__author__ = "Andrew Lee, Jaffer Ghouse"


_log = idaeslog.getLogger(__name__)

# TODO : Custom terms in material balances, other types of material balances
# Diffusion terms need to be added


# Enumerate options for area
class DistributedVars(Enum):
    """
    Enum indicating if a variable is constant across the spatial domain.
    """

    variant = 0
    uniform = 1


@declare_process_block_class(
    "ControlVolume1DBlock",
    doc="""
    ControlVolume1DBlock is a specialized Pyomo block for IDAES control volume
    blocks discretized in one spatial direction, and contains instances of
    ControlVolume1DBlockData.

    ControlVolume1DBlock should be used for any control volume with a defined
    volume and distinct inlets and outlets where there is a single spatial
    domain parallel to the material flow direction. This encompasses unit
    operations such as plug flow reactors and pipes.""",
)
class ControlVolume1DBlockData(ControlVolumeBlockData):
    """
    1-Dimensional ControlVolume Class

    This class forms the core of all 1-D IDAES models. It provides
    methods to build property and reaction blocks, and add mass, energy and
    momentum balances. The form of the terms used in these constraints is
    specified in the chosen property package.
    """

    CONFIG = ControlVolumeBlockData.CONFIG()
    CONFIG.declare(
        "area_definition",
        ConfigValue(
            default=DistributedVars.uniform,
            domain=In(DistributedVars),
            description="Argument for defining form of area variable",
            doc="""Argument defining whether area variable should be spatially
        variant or not. **default** - DistributedVars.uniform.
        **Valid values:** {
        DistributedVars.uniform - area does not vary across spatial domain,
        DistributedVars.variant - area can vary over the domain and is indexed
        by time and space.}""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default=None,
            domain=is_transformation_method,
            description="DAE transformation method",
            doc="""Method to use to transform domain. Must be a method recognised
by the Pyomo TransformationFactory.""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=None,
            domain=is_transformation_scheme,
            description="DAE transformation scheme",
            doc="""Scheme to use when transforming domain. See Pyomo
documentation for supported schemes.""",
        ),
    )
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=None,
            domain=int,
            description="Number of finite elements",
            doc="""Number of finite elements to use in transformation (equivalent
to Pyomo nfe argument).""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=None,
            domain=int,
            description="Number of collocation points",
            doc="""Number of collocation points to use (equivalent to Pyomo ncp
argument).""",
        ),
    )

    def build(self):
        """
        Build method for ControlVolume1DBlock blocks.

        Returns:
            None
        """
        # Call build method from base class
        super(ControlVolume1DBlockData, self).build()

        self._validate_config_args()

    def _validate_config_args(self):
        # Validate DAE config arguments
        if self.config.transformation_method is None:
            raise ConfigurationError(
                "{} was not provided a value for the transformation_method"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )

        if self.config.transformation_scheme is None:
            raise ConfigurationError(
                "{} was not provided a value for the transformation_scheme"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )
        elif (
            self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme not in ["BACKWARD", "FORWARD"]
        ) or (
            self.config.transformation_method == "dae.collocation"
            and self.config.transformation_scheme
            not in ["LAGRANGE-LEGENDRE", "LAGRANGE-RADAU"]
        ):
            raise ConfigurationError(
                "{} transformation_scheme configuration argument is not "
                "consistent with transformation_method argument. See Pyomo"
                " documentation for argument options.".format(self.name)
            )

    def add_geometry(
        self,
        length_domain=None,
        length_domain_set=None,
        length_var=None,
        flow_direction=FlowDirection.forward,
    ):
        """
        Method to create spatial domain and volume Var in ControlVolume.

        Args:
            length_domain: (optional) a ContinuousSet to use as the length
                domain for the ControlVolume. If not provided, a new
                ContinuousSet will be created (default=None). ContinuousSet
                should be normalized to run between 0 and 1.
            length_domain_set: (optional) list of point to use to initialize
                a new ContinuousSet if length_domain is not provided (default = [0.0, 1.0]).
            length_var: (optional) external variable to use for the length of
                the spatial domain. If a variable is provided, a reference
                will be made to this in place of the length Var.
            flow_direction: argument indicating direction of material flow
                relative to length domain. Valid values:

                    * FlowDirection.forward (default), flow goes from 0 to 1.
                    * FlowDirection.backward, flow goes from 1 to 0

        Returns:
            None
        """
        units = self.config.property_package.get_metadata().get_derived_units

        if length_domain is not None:
            # Validate domain and make a reference
            if isinstance(length_domain, ContinuousSet):
                add_object_reference(self, "length_domain", length_domain)
            else:
                raise ConfigurationError(
                    "{} length_domain argument must be a Pyomo "
                    "ContinuousSet object".format(self.name)
                )
        else:
            # Create new length domain
            if length_domain_set is None:
                length_domain_set = [0.0, 1.0]

            self.length_domain = ContinuousSet(
                bounds=(0.0, 1.0),
                initialize=length_domain_set,
                doc="Normalized length domain",
            )

        # Validated and create flow direction attribute
        if flow_direction in (flwd for flwd in FlowDirection):
            self._flow_direction = flow_direction
        else:
            raise ConfigurationError(
                "{} invalid value for flow_direction "
                "argument. Must be a FlowDirection Enum.".format(self.name)
            )
        if flow_direction is FlowDirection.forward:
            self._flow_direction_term = -1
        else:
            self._flow_direction_term = 1

        # Add geometry variables and constraints
        if self.config.area_definition == DistributedVars.variant:
            self.area = Var(
                self.flowsheet().time,
                self.length_domain,
                initialize=1.0,
                doc="Cross-sectional area of Control Volume",
                units=units("area"),
            )
        else:
            self.area = Var(
                initialize=1.0,
                doc="Cross-sectional area of Control Volume",
                units=units("area"),
            )

        if length_var is not None:
            # Validate length_Var and add a reference
            if not isinstance(length_var, (Var, Param, Expression)):
                raise ConfigurationError(
                    f"{self.name} length_var must be a Pyomo Var, Param or "
                    "Expression."
                )
            elif length_var.is_indexed():
                raise ConfigurationError(
                    f"{self.name} length_var must be a scalar (unindexed) component."
                )
            add_object_reference(self, "length", length_var)
        else:
            self.length = Var(
                initialize=1.0, doc="Length of Control Volume", units=units("length")
            )

    def add_state_blocks(
        self, information_flow=FlowDirection.forward, has_phase_equilibrium=None
    ):
        """
        This method constructs the state blocks for the control volume.

        Args:
            information_flow: a FlowDirection Enum indicating whether
                information flows from inlet-to-outlet or outlet-to-inlet
            has_phase_equilibrium: indicates whether equilibrium calculations
                will be required in state blocks
            package_arguments: dict-like object of arguments to be passed to
                state blocks as construction arguments

        Returns:
            None
        """
        if has_phase_equilibrium is None:
            raise ConfigurationError(
                "{} add_state_blocks method was not provided with a "
                "has_phase_equilibrium argument.".format(self.name)
            )
        elif has_phase_equilibrium not in [True, False]:
            raise ConfigurationError(
                "{} add_state_blocks method was provided with an invalid "
                "has_phase_equilibrium argument. Must be True or False".format(
                    self.name
                )
            )

        # d0 is config for defined state d1 is config for not defined state
        d0 = dict(**self.config.property_package_args)
        d0.update(has_phase_equilibrium=has_phase_equilibrium, defined_state=True)
        d1 = copy.copy(d0)
        d1["defined_state"] = False

        def idx_map(i):  # i = (t, x)
            if (
                information_flow == FlowDirection.forward
                and i[1] == self.length_domain.first()
            ):
                return 0
            elif (
                information_flow == FlowDirection.backward
                and i[1] == self.length_domain.last()
            ):
                return 0
            else:
                return 1

        self.properties = self.config.property_package.build_state_block(
            self.flowsheet().time,
            self.length_domain,
            doc="Material properties",
            initialize={0: d0, 1: d1},  # TODO: What if the domain has different bounds?
            idx_map=idx_map,
        )

    def add_reaction_blocks(self, has_equilibrium=None):
        """
        This method constructs the reaction block for the control volume.

        Args:
            has_equilibrium: indicates whether equilibrium calculations
                will be required in reaction block
            package_arguments: dict-like object of arguments to be passed to
                reaction block as construction arguments

        Returns:
            None
        """
        if has_equilibrium is None:
            raise ConfigurationError(
                "{} add_reaction_blocks method was not provided with a "
                "has_equilibrium argument.".format(self.name)
            )
        elif has_equilibrium not in [True, False]:
            raise ConfigurationError(
                "{} add_reaction_blocks method was provided with an "
                "invalid has_equilibrium argument. Must be True or False".format(
                    self.name
                )
            )

        # TODO : Should not have ReactionBlock at inlet
        tmp_dict = dict(**self.config.reaction_package_args)
        tmp_dict["state_block"] = self.properties
        tmp_dict["has_equilibrium"] = has_equilibrium

        self.reactions = self.config.reaction_package.build_reaction_block(
            self.flowsheet().time,
            self.length_domain,
            doc="Reaction properties in control volume",
            **tmp_dict,
        )  # TODO: Do we need something similar to above to skip equilibrium at bounds?

    def _add_material_balance_common(
        self,
        balance_type,
        has_rate_reactions,
        has_equilibrium_reactions,
        has_phase_equilibrium,
        has_mass_transfer,
        custom_molar_term,
        custom_mass_term,
    ):
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        component_list = self.properties.component_list
        phase_list = self.properties.phase_list
        pc_set = self.properties.phase_component_set

        # Check that reaction block exists if required
        if has_rate_reactions or has_equilibrium_reactions:
            try:
                rblock = self.reactions
            except AttributeError:
                raise ConfigurationError(
                    "{} does not contain a Reaction Block, but material "
                    "balances have been set to contain reaction terms. "
                    "Please construct a reaction block before adding "
                    "balance equations.".format(self.name)
                )

        if has_equilibrium_reactions:
            # Check that reaction block is set to calculate equilibrium
            for t in self.flowsheet().time:
                for x in self.length_domain:
                    if self.reactions[t, x].config.has_equilibrium is False:
                        raise ConfigurationError(
                            "{} material balance was set to include "
                            "equilibrium reactions, however the associated "
                            "ReactionBlock was not set to include equilibrium "
                            "constraints (has_equilibrium_reactions=False). "
                            "Please correct your configuration arguments.".format(
                                self.name
                            )
                        )

        if has_phase_equilibrium:
            # First, check that phase equilibrium makes sense
            if len(self.config.property_package.phase_list) < 2:
                msg = (
                    "Property package has only one phase; control volume cannot include phase "
                    "equilibrium terms. Some property packages support phase equilibrium "
                    "implicitly in which case additional terms are not necessary. "
                    "You should set has_phase_equilibrium=False."
                )
                deprecation_warning(
                    msg=msg, logger=_log, version="2.0.0", remove_in="3.0.0"
                )
                has_phase_equilibrium = False
            else:
                # Check that state blocks are set to calculate equilibrium
                for t in self.flowsheet().time:
                    for x in self.length_domain:
                        if not self.properties[t, x].config.has_phase_equilibrium:
                            raise ConfigurationError(
                                "{} material balance was set to include phase "
                                "equilibrium, however the associated "
                                "StateBlock was not set to include equilibrium "
                                "constraints (has_phase_equilibrium=False). Please"
                                " correct your configuration arguments.".format(
                                    self.name
                                )
                            )

        # Get units from property package
        units = self.config.property_package.get_metadata().get_derived_units

        if units("length") is not None:
            if (
                self.properties[
                    self.flowsheet().time.first(), self.length_domain.first()
                ].get_material_flow_basis()
                == MaterialFlowBasis.molar
            ):
                holdup_l_units = units("amount") / units("length")
                flow_units = units("flow_mole")
                flow_l_units = units("flow_mole") / units("length")
            elif (
                self.properties[
                    self.flowsheet().time.first(), self.length_domain.first()
                ].get_material_flow_basis()
                == MaterialFlowBasis.mass
            ):
                holdup_l_units = units("mass") / units("length")
                flow_units = units("flow_mass")
                flow_l_units = units("flow_mass") / units("length")
            else:
                holdup_l_units = None
                flow_units = None
                flow_l_units = None
        else:
            holdup_l_units = None
            flow_units = None
            flow_l_units = None

        # Get units for accumulation term if required
        if self.config.dynamic:
            f_time_units = self.flowsheet().time_units
            if (f_time_units is None) ^ (units("time") is None):
                raise ConfigurationError(
                    "{} incompatible time unit specification between "
                    "flowsheet and property package. Either both must use "
                    "units, or neither.".format(self.name)
                )

            if f_time_units is None:
                acc_units = None
            elif (
                self.properties[
                    self.flowsheet().time.first(), self.length_domain.first()
                ].get_material_flow_basis()
                == MaterialFlowBasis.other
            ):
                acc_units = None
            else:
                acc_units = holdup_l_units / f_time_units

        # Check if reaction package exists, and get units
        if hasattr(self.config, "reaction_package"):
            if self.config.reaction_package is not None:
                if (
                    self.reactions[
                        self.flowsheet().time.first(), self.length_domain.first()
                    ].get_reaction_rate_basis()
                    == MaterialFlowBasis.molar
                ):
                    rxn_flow_l_units = units("flow_mole") / units("length")
                elif (
                    self.reactions[
                        self.flowsheet().time.first(), self.length_domain.first()
                    ].get_reaction_rate_basis()
                    == MaterialFlowBasis.mass
                ):
                    rxn_flow_l_units = units("flow_mass") / units("length")
                else:  # reaction basis not defined
                    rxn_flow_l_units = None
            else:  # reaction package is NoneType object
                rxn_flow_l_units = None
        else:  # reaction package not defined
            rxn_flow_l_units = None

        # Material holdup and accumulation
        if has_holdup:
            self.material_holdup = Var(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                domain=Reals,
                initialize=1.0,
                doc="Material holdup per unit length",
                units=holdup_l_units,
            )
        if dynamic:
            self.material_accumulation = DerivativeVar(
                self.material_holdup,
                wrt=self.flowsheet().time,
                doc="Material accumulation per unit length",
                units=acc_units,
            )

        # Create material balance terms as required
        # Flow terms and derivatives
        self._flow_terms = Var(
            self.flowsheet().time,
            self.length_domain,
            pc_set,
            initialize=1.0,
            doc="Flow terms for material balance equations",
            units=flow_units,
        )

        @self.Constraint(
            self.flowsheet().time,
            self.length_domain,
            pc_set,
            doc="Material flow linking constraints",
        )
        def material_flow_linking_constraints(b, t, x, p, j):
            return b._flow_terms[t, x, p, j] == b.properties[
                t, x
            ].get_material_flow_terms(p, j)

        self.material_flow_dx = DerivativeVar(
            self._flow_terms,
            wrt=self.length_domain,
            doc="Partial derivative of material flow wrt to normalized length",
            units=flow_units,
        )

        # Kinetic reaction generation
        if has_rate_reactions:
            if not hasattr(self.config.reaction_package, "rate_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of rate "
                    "reactions (rate_reaction_idx), thus does not support "
                    "rate-based reactions.".format(self.name)
                )
            self.rate_reaction_generation = Var(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated in "
                "by kinetic reactions per unit length",
                units=rxn_flow_l_units,
            )  # use reaction package flow basis

        # Equilibrium reaction generation
        if has_equilibrium_reactions:
            if not hasattr(self.config.reaction_package, "equilibrium_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Reaction package does not contain a list of "
                    "equilibrium reactions (equilibrium_reaction_idx), thus "
                    "does not support equilibrium-based reactions.".format(self.name)
                )
            self.equilibrium_reaction_generation = Var(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated by equilibrium "
                "reactions per unit length",
                units=rxn_flow_l_units,
            )  # use reaction package flow basis

        # Inherent reaction generation
        if self.properties.include_inherent_reactions:
            if not hasattr(self.config.property_package, "inherent_reaction_idx"):
                raise PropertyNotSupportedError(
                    "{} Property package does not contain a list of "
                    "inherent reactions (inherent_reaction_idx), but "
                    "include_inherent_reactions is True.".format(self.name)
                )
            self.inherent_reaction_generation = Var(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Amount of component generated in control volume "
                "by inherent reactions",
                units=flow_l_units,
            )  # use property package flow basis

        # Phase equilibrium generation
        if has_phase_equilibrium and balance_type == MaterialBalanceType.componentPhase:
            if not hasattr(self.config.property_package, "phase_equilibrium_idx"):
                raise PropertyNotSupportedError(
                    "{} Property package does not contain a list of phase "
                    "equilibrium reactions (phase_equilibrium_idx), thus does "
                    "not support phase equilibrium.".format(self.name)
                )
            self.phase_equilibrium_generation = Var(
                self.flowsheet().time,
                self.length_domain,
                self.config.property_package.phase_equilibrium_idx,
                domain=Reals,
                initialize=0.0,
                doc="Amount of generation in unit by phase "
                "equilibria per unit length",
                units=flow_l_units,
            )  # use property package flow basis

        # Material transfer term
        if has_mass_transfer:
            self.mass_transfer_term = Var(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Component material transfer into unit per unit length",
                units=flow_l_units,
            )

        # Create rules to substitute material balance terms
        def phase_equilibrium_term(b, t, x, p, j):
            if (
                has_phase_equilibrium
                and balance_type == MaterialBalanceType.componentPhase
            ):
                sd = {}
                for r in b.config.property_package.phase_equilibrium_idx:
                    if b.config.property_package.phase_equilibrium_list[r][0] == j:
                        if (
                            b.config.property_package.phase_equilibrium_list[r][1][0]
                            == p
                        ):
                            sd[r] = 1
                        elif (
                            b.config.property_package.phase_equilibrium_list[r][1][1]
                            == p
                        ):
                            sd[r] = -1
                        else:
                            sd[r] = 0
                    else:
                        sd[r] = 0

                return sum(
                    b.phase_equilibrium_generation[t, x, r] * sd[r]
                    for r in b.config.property_package.phase_equilibrium_idx
                )
            else:
                return 0

        # TODO: Need to set material_holdup = 0 for non-present component-phase
        # pairs. Not ideal, but needed to close DoF. Is there a better way?

        # Material Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                doc="Material holdup calculations",
            )
            def material_holdup_calculation(b, t, x, p, j):
                if (p, j) in pc_set:
                    return b.material_holdup[t, x, p, j] == (
                        b._area_func(t, x)
                        * self.phase_fraction[t, x, p]
                        * b.properties[t, x].get_material_density_terms(p, j)
                    )

        if has_rate_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.rate_reaction_extent = Var(
                self.flowsheet().time,
                self.length_domain,
                self.config.reaction_package.rate_reaction_idx,
                domain=Reals,
                initialize=0.0,
                doc="Extent of kinetic reactions at point x",
                units=rxn_flow_l_units,
            )  # use reaction package flow basis

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                doc="Kinetic reaction stoichiometry constraint",
            )
            def rate_reaction_stoichiometry_constraint(b, t, x, p, j):
                if (
                    b.config.transformation_scheme != "FORWARD"
                    and x == b.length_domain.first()
                ) or (
                    b.config.transformation_scheme == "FORWARD"
                    and x == b.length_domain.last()
                ):
                    return Constraint.Skip
                elif (p, j) in pc_set:
                    rparam = rblock[t, x].config.parameters
                    return b.rate_reaction_generation[t, x, p, j] == (
                        sum(
                            rparam.rate_reaction_stoichiometry[r, p, j]
                            * b.rate_reaction_extent[t, x, r]
                            for r in b.config.reaction_package.rate_reaction_idx
                        )
                    )
                else:
                    return Constraint.Skip

        if has_equilibrium_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.equilibrium_reaction_extent = Var(
                self.flowsheet().time,
                self.length_domain,
                self.config.reaction_package.equilibrium_reaction_idx,
                domain=Reals,
                initialize=0.0,
                doc="Extent of equilibrium reactions at point x",
                units=rxn_flow_l_units,
            )  # use reaction package flow basis

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                doc="Equilibrium reaction stoichiometry",
            )
            def equilibrium_reaction_stoichiometry_constraint(b, t, x, p, j):
                if (
                    b.config.transformation_scheme != "FORWARD"
                    and x == b.length_domain.first()
                ) or (
                    b.config.transformation_scheme == "FORWARD"
                    and x == b.length_domain.last()
                ):
                    return Constraint.Skip
                elif (p, j) in pc_set:
                    return b.equilibrium_reaction_generation[t, x, p, j] == (
                        sum(
                            rblock[
                                t, x
                            ].config.parameters.equilibrium_reaction_stoichiometry[
                                r, p, j
                            ]
                            * b.equilibrium_reaction_extent[t, x, r]
                            for r in b.config.reaction_package.equilibrium_reaction_idx
                        )
                    )
                else:
                    return Constraint.Skip

        if self.properties.include_inherent_reactions:
            # Add extents of reaction and stoichiometric constraints
            self.inherent_reaction_extent = Var(
                self.flowsheet().time,
                self.length_domain,
                self.config.property_package.inherent_reaction_idx,
                domain=Reals,
                initialize=0.0,
                doc="Extent of inherent reactions at point x",
                units=flow_l_units,
            )  # use property package flow basis

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                doc="Inherent reaction stoichiometry",
            )
            def inherent_reaction_stoichiometry_constraint(b, t, x, p, j):
                if (
                    b.config.transformation_scheme != "FORWARD"
                    and x == b.length_domain.first()
                ) or (
                    b.config.transformation_scheme == "FORWARD"
                    and x == b.length_domain.last()
                ):
                    return Constraint.Skip
                elif (p, j) in pc_set:
                    return b.inherent_reaction_generation[t, x, p, j] == (
                        sum(
                            b.properties[
                                t, x
                            ].config.parameters.inherent_reaction_stoichiometry[r, p, j]
                            * b.inherent_reaction_extent[t, x, r]
                            for r in b.config.property_package.inherent_reaction_idx
                        )
                    )
                else:
                    return Constraint.Skip

        # Add custom terms and material balances
        if balance_type == MaterialBalanceType.componentPhase:

            def user_term_mol(b, t, x, p, j):
                flow_basis = b.properties[t, x].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.molar:
                    return custom_molar_term(t, x, p, j)
                elif flow_basis == MaterialFlowBasis.mass:
                    try:
                        return (
                            custom_molar_term(t, x, p, j)
                            * b.properties[t, x].mw_comp[j]
                        )
                    except AttributeError:
                        raise PropertyNotSupportedError(
                            "{} property package does not support "
                            "molecular weight (mw), which is required for "
                            "using custom terms in material balances.".format(self.name)
                        )
                else:
                    raise ConfigurationError(
                        "{} contained a custom_molar_term argument, but "
                        "the property package used an undefined basis "
                        "(MaterialFlowBasis.other). Custom terms can "
                        "only be used when the property package declares "
                        "a molar or mass flow basis.".format(self.name)
                    )

            def user_term_mass(b, t, x, p, j):
                flow_basis = b.properties[t, x].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.mass:
                    return custom_mass_term(t, x, p, j)
                elif flow_basis == MaterialFlowBasis.molar:
                    try:
                        return (
                            custom_mass_term(t, x, p, j) / b.properties[t, x].mw_comp[j]
                        )
                    except AttributeError:
                        raise PropertyNotSupportedError(
                            "{} property package does not support "
                            "molecular weight (mw), which is required for "
                            "using custom terms in material balances.".format(self.name)
                        )
                else:
                    raise ConfigurationError(
                        "{} contained a custom_mass_term argument, but "
                        "the property package used an undefined basis "
                        "(MaterialFlowBasis.other). Custom terms can "
                        "only be used when the property package declares "
                        "a molar or mass flow basis.".format(self.name)
                    )

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                pc_set,
                doc="Material balances",
            )
            def material_balances(b, t, x, p, j):
                if (
                    b.config.transformation_scheme != "FORWARD"
                    and x == b.length_domain.first()
                ) or (
                    b.config.transformation_scheme == "FORWARD"
                    and x == b.length_domain.last()
                ):
                    return Constraint.Skip
                else:
                    if (p, j) in pc_set:
                        if dynamic:
                            lhs = b.length * pyunits.convert(
                                b.material_accumulation[t, x, p, j],
                                to_units=flow_l_units,
                            )
                        else:
                            lhs = 0

                        rhs = b._flow_direction_term * b.material_flow_dx[t, x, p, j]

                        if has_rate_reactions:
                            rhs += (
                                b.length
                                * b.rate_reaction_generation[t, x, p, j]
                                * b._rxn_rate_conv(t, x, j)
                            )

                        if has_equilibrium_reactions:
                            rhs += (
                                b.length * b.equilibrium_reaction_generation[t, x, p, j]
                            )

                        if b.properties.include_inherent_reactions:
                            rhs += b.length * b.inherent_reaction_generation[t, x, p, j]

                        if (
                            has_phase_equilibrium
                            and balance_type == MaterialBalanceType.componentPhase
                        ):
                            rhs += b.length * phase_equilibrium_term(b, t, x, p, j)

                        if has_mass_transfer:
                            rhs += b.length * b.mass_transfer_term[t, x, p, j]

                        if custom_molar_term is not None:
                            rhs += b.length * user_term_mol(b, t, x, p, j)

                        if custom_mass_term is not None:
                            rhs += b.length * user_term_mass(b, t, x, p, j)

                        return lhs == rhs

                    else:
                        return Constraint.Skip

        elif balance_type == MaterialBalanceType.componentTotal:

            def user_term_mol(b, t, x, j):
                flow_basis = b.properties[t, x].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.molar:
                    return custom_molar_term(t, x, j)
                elif flow_basis == MaterialFlowBasis.mass:
                    try:
                        return (
                            custom_molar_term(t, x, j) * b.properties[t, x].mw_comp[j]
                        )
                    except AttributeError:
                        raise PropertyNotSupportedError(
                            "{} property package does not support "
                            "molecular weight (mw), which is required for "
                            "using custom terms in material balances.".format(self.name)
                        )
                else:
                    raise ConfigurationError(
                        "{} contained a custom_molar_term argument, but "
                        "the property package used an undefined basis "
                        "(MaterialFlowBasis.other). Custom terms can "
                        "only be used when the property package declares "
                        "a molar or mass flow basis.".format(self.name)
                    )

            def user_term_mass(b, t, x, j):
                flow_basis = b.properties[t, x].get_material_flow_basis()
                if flow_basis == MaterialFlowBasis.mass:
                    return custom_mass_term(t, x, j)
                elif flow_basis == MaterialFlowBasis.molar:
                    try:
                        return custom_mass_term(t, x, j) / b.properties[t, x].mw_comp[j]
                    except AttributeError:
                        raise PropertyNotSupportedError(
                            "{} property package does not support "
                            "molecular weight (mw), which is required for "
                            "using custom terms in material balances.".format(self.name)
                        )
                else:
                    raise ConfigurationError(
                        "{} contained a custom_mass_term argument, but "
                        "the property package used an undefined basis "
                        "(MaterialFlowBasis.other). Custom terms can "
                        "only be used when the property package declares "
                        "a molar or mass flow basis.".format(self.name)
                    )

            # Add component balances
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                component_list,
                doc="Material balances",
            )
            def material_balances(b, t, x, j):
                if (
                    b.config.transformation_scheme != "FORWARD"
                    and x == b.length_domain.first()
                ) or (
                    b.config.transformation_scheme == "FORWARD"
                    and x == b.length_domain.last()
                ):
                    return Constraint.Skip
                else:
                    cplist = []
                    for p in phase_list:
                        if (p, j) in pc_set:
                            cplist.append(p)

                    if dynamic:
                        lhs = b.length * pyunits.convert(
                            sum(b.material_accumulation[t, x, p, j] for p in cplist),
                            to_units=flow_l_units,
                        )
                    else:
                        lhs = 0

                    rhs = b._flow_direction_term * sum(
                        b.material_flow_dx[t, x, p, j] for p in cplist
                    )

                    if has_rate_reactions:
                        rhs += (
                            b.length
                            * sum(
                                b.rate_reaction_generation[t, x, p, j] for p in cplist
                            )
                            * b._rxn_rate_conv(t, x, j)
                        )

                    if has_equilibrium_reactions:
                        rhs += b.length * sum(
                            b.equilibrium_reaction_generation[t, x, p, j]
                            for p in cplist
                        )

                    if b.properties.include_inherent_reactions:
                        rhs += b.length * sum(
                            b.inherent_reaction_generation[t, x, p, j] for p in cplist
                        )

                    if has_mass_transfer:
                        rhs += b.length * sum(
                            b.mass_transfer_term[t, x, p, j] for p in cplist
                        )

                    if custom_molar_term is not None:
                        rhs += b.length * user_term_mol(b, t, x, j)

                    if custom_mass_term is not None:
                        rhs += b.length * user_term_mass(b, t, x, j)

                    return lhs == rhs

    def add_phase_component_balances(
        self,
        has_rate_reactions=False,
        has_equilibrium_reactions=False,
        has_phase_equilibrium=False,
        has_mass_transfer=False,
        custom_molar_term=None,
        custom_mass_term=None,
    ):
        """
        This method constructs a set of 1D material balances indexed by time,
        length, phase and component.

        Args:
            has_rate_reactions: whether default generation terms for rate
                reactions should be included in material balances
            has_equilibrium_reactions: whether generation terms should for
                chemical equilibrium reactions should be included in
                material balances
            has_phase_equilibrium: whether generation terms should for phase
                equilibrium behaviour should be included in material
                balances
            has_mass_transfer: whether generic mass transfer terms should be
                included in material balances
            custom_molar_term: a Pyomo Expression representing custom terms to
                be included in material balances on a molar basis. Expression
                must be indexed by time, length domain, phase list and component list
            custom_mass_term: a Pyomo Expression representing custom terms to
                be included in material balances on a mass basis. Expression must
                be indexed by time, length domain, phase list and component list

        Returns:
            Constraint object representing material balances
        """
        self._add_material_balance_common(
            balance_type=MaterialBalanceType.componentPhase,
            has_rate_reactions=has_rate_reactions,
            has_equilibrium_reactions=has_equilibrium_reactions,
            has_phase_equilibrium=has_phase_equilibrium,
            has_mass_transfer=has_mass_transfer,
            custom_molar_term=custom_molar_term,
            custom_mass_term=custom_mass_term,
        )

        return self.material_balances

    def add_total_component_balances(
        self,
        has_rate_reactions=False,
        has_equilibrium_reactions=False,
        has_phase_equilibrium=False,
        has_mass_transfer=False,
        custom_molar_term=None,
        custom_mass_term=None,
    ):
        """
        This method constructs a set of 1D material balances indexed by time
        length and component.

        Args:
            has_rate_reactions: whether default generation terms for rate
                reactions should be included in material balances
            has_equilibrium_reactions: whether generation terms should for
                chemical equilibrium reactions should be included in
                material balances
            has_phase_equilibrium: whether generation terms should for phase
                equilibrium behaviour should be included in material
                balances
            has_mass_transfer: whether generic mass transfer terms should be
                included in material balances
            custom_molar_term: a Pyomo Expression representing custom terms to
                be included in material balances on a molar basis.
                Expression must be indexed by time, length domain and
                component list
            custom_mass_term: a Pyomo Expression representing custom terms to
                be included in material balances on a mass basis.
                Expression must be indexed by time, length domain and
                component list

        Returns:
            Constraint object representing material balances
        """
        self._add_material_balance_common(
            balance_type=MaterialBalanceType.componentTotal,
            has_rate_reactions=has_rate_reactions,
            has_equilibrium_reactions=has_equilibrium_reactions,
            has_phase_equilibrium=has_phase_equilibrium,
            has_mass_transfer=has_mass_transfer,
            custom_molar_term=custom_molar_term,
            custom_mass_term=custom_mass_term,
        )

        return self.material_balances

    def add_total_element_balances(
        self,
        has_rate_reactions=False,
        has_equilibrium_reactions=False,
        has_phase_equilibrium=False,
        has_mass_transfer=False,
        custom_elemental_term=None,
    ):
        """
        This method constructs a set of 1D element balances indexed by time and
        length.

        Args:
            has_rate_reactions: whether default generation terms for rate
                reactions should be included in material balances
            has_equilibrium_reactions: whether generation terms should for
                chemical equilibrium reactions should be included in
                material balances
            has_phase_equilibrium: whether generation terms should for phase
                equilibrium behaviour should be included in material balances
            has_mass_transfer: whether generic mass transfer terms should be
                included in material balances
            custom_elemental_term: a Pyomo Expression representing custom
                terms to be included in material balances on a molar
                elemental basis. Expression must be indexed by time, length
                and element list

        Returns:
            Constraint object representing material balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        component_list = self.properties.component_list
        pc_set = self.properties.phase_component_set

        # Check that property package supports element balances
        if not hasattr(self.config.property_package, "element_list"):
            raise PropertyNotSupportedError(
                "{} property package provided does not contain a list of "
                "elements (element_list), and thus does not support "
                "elemental material balances. Please choose another type "
                "of material balance or a property package which supports "
                "elemental balances."
            )

        # Check validity of arguments to write the total elemental balance
        if has_rate_reactions:
            raise ConfigurationError(
                "{} add_total_element_balances method as provided with "
                "argument has_rate_reactions = True. Total element "
                "balances do not support rate based reactions, "
                "please correct your configuration arguments".format(self.name)
            )

        if has_equilibrium_reactions:
            raise ConfigurationError(
                "{} add_total_element_balances method as provided with "
                "argument has_equilibrium_reactions = True. Total element "
                "balances do not support equilibrium based reactions, "
                "please correct your configuration arguments".format(self.name)
            )

        if has_phase_equilibrium:
            # Check that state blocks are set to calculate equilibrium
            raise ConfigurationError(
                "{} add_total_element_balances method as provided with "
                "argument has_phase_equilibrium = True. Total element "
                "balances do not support equilibrium based reactions, "
                "please correct your configuration arguments".format(self.name)
            )

        # Get units from property package
        units = self.config.property_package.get_metadata().get_derived_units

        if units("amount") is not None:
            amount_l_units = units("amount") / units("length")
            flow_units = units("flow_mole")
            flow_l_units = units("flow_mole") / units("length")
        else:
            amount_l_units = None
            flow_units = None
            flow_l_units = None

        # Get units for accumulation term if required
        if self.config.dynamic:
            f_time_units = self.flowsheet().time_units
            if (f_time_units is None) ^ (units("time") is None):
                raise ConfigurationError(
                    "{} incompatible time unit specification between "
                    "flowsheet and property package. Either both must use "
                    "units, or neither.".format(self.name)
                )

            if f_time_units is None:
                acc_units = None
            else:
                acc_units = amount_l_units / f_time_units

        # Identify linearly dependent elements
        # It is possible for there to be linearly dependent element balances
        # e.g. if a single species is the only source of two different elements
        linearly_dependent = []

        # Get a representative time point
        rtime = self.flowsheet().time.first()
        rdomain = self.length_domain.first()

        # For each component in the material, search for elements which are
        # unique to it
        for i in component_list:
            unique_elements = []
            for e in self.config.property_package.element_list:
                if self.properties[rtime, rdomain].params.element_comp[i][e] != 0:
                    # Assume unique until shown otherwise
                    unique = True

                    for j in component_list:
                        if j == i:
                            continue

                        # If element appears in any other component, not unique
                        if (
                            self.properties[rtime, rdomain].params.element_comp[j][e]
                            != 0
                        ):
                            unique = False

                    if unique:
                        unique_elements.append(e)

            # If more than 1 unique element, they are linearly dependent
            if len(unique_elements) > 1:
                # Add all but the first to the list of linearly dependent
                linearly_dependent.extend(unique_elements[1:])

        # Set indexing set for element balances
        if len(linearly_dependent) == 0:
            # No linearly depednet equations, so use full element list
            e_index = self.config.property_package.element_list
        else:
            # Otherwise, use only non-dependent elements, and log a message
            _log.info_low(
                "{} detected linearly dependent element balance "
                "equations. Element balances will NOT be written "
                "for the following elements: {}".format(self.name, linearly_dependent)
            )
            e_index = self.config.property_package.element_list - linearly_dependent

        # Add Material Balance terms
        if has_holdup:
            self.element_holdup = Var(
                self.flowsheet().time,
                self.length_domain,
                self.config.property_package.element_list,
                domain=Reals,
                initialize=1.0,
                doc="Elemental holdup per unit length",
                units=amount_l_units,
            )

        if dynamic:
            self.element_accumulation = DerivativeVar(
                self.element_holdup,
                wrt=self.flowsheet().time,
                doc="Elemental accumulation per unit length",
                units=acc_units,
            )

        self.elemental_flow_term = Var(
            self.flowsheet().time,
            self.length_domain,
            self.config.property_package.element_list,
            initialize=1.0,
            doc="Elemental flow terms",
            units=flow_units,
        )

        # Method to convert mass flow basis to mole flow basis
        def conv_factor(b, t, x, j):
            flow_basis = b.properties[t, x].get_material_flow_basis()
            if flow_basis == MaterialFlowBasis.molar:
                return 1
            elif flow_basis == MaterialFlowBasis.mass:
                return 1 / b.properties[t, x].mw_comp[j]
            else:
                raise BalanceTypeNotSupportedError(
                    "{} property package MaterialFlowBasis == 'other'. Cannot "
                    "automatically generate elemental balances.".format(self.name)
                )

        @self.Constraint(
            self.flowsheet().time,
            self.length_domain,
            self.config.property_package.element_list,
            doc="Elemental flow constraints",
        )
        def elemental_flow_constraint(b, t, x, e):
            return b.elemental_flow_term[t, x, e] == (
                sum(
                    conv_factor(b, t, x, j)
                    * b.properties[t, x].get_material_flow_terms(p, j)
                    * b.properties[t, x].config.parameters.element_comp[j][e]
                    for p, j in pc_set
                )
            )

        self.elemental_flow_dx = DerivativeVar(
            self.elemental_flow_term,
            wrt=self.length_domain,
            doc="Partial derivative of elemental flow wrt normalized length",
            units=flow_units,
        )

        # Create material balance terms as needed
        if has_mass_transfer:
            self.elemental_mass_transfer_term = Var(
                self.flowsheet().time,
                self.length_domain,
                e_index,
                domain=Reals,
                initialize=0.0,
                doc="Element material transfer into unit per unit length",
                units=flow_l_units,
            )

        # Element balances
        @self.Constraint(
            self.flowsheet().time,
            self.length_domain,
            e_index,
            doc="Elemental material balances",
        )
        def element_balances(b, t, x, e):
            if (
                b.config.transformation_scheme != "FORWARD"
                and x == b.length_domain.first()
            ) or (
                b.config.transformation_scheme == "FORWARD"
                and x == b.length_domain.last()
            ):
                return Constraint.Skip
            else:
                if dynamic:
                    lhs = b.length * pyunits.convert(
                        b.element_accumulation[t, x, e], to_units=flow_l_units
                    )
                else:
                    lhs = 0
                rhs = b._flow_direction_term * b.elemental_flow_dx[t, x, e]

                if has_mass_transfer:
                    rhs += b.length * b.elemental_mass_transfer_term[t, x, e]

                if custom_elemental_term is not None:
                    rhs += b.length * custom_elemental_term(t, x, e)

                # TODO : Add diffusion terms

                return lhs == rhs

        # Elemental Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                self.config.property_package.element_list,
                doc="Elemental holdup calculation",
            )
            def elemental_holdup_calculation(b, t, x, e):
                return b.element_holdup[t, x, e] == (
                    b._area_func(t, x)
                    * sum(
                        conv_factor(b, t, x, j)
                        * b.phase_fraction[t, x, p]
                        * b.properties[t, x].get_material_density_terms(p, j)
                        * b.properties[t, x].config.parameters.element_comp[j][e]
                        for p, j in pc_set
                    )
                )

        return self.element_balances

    def add_total_material_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} 1D control volumes do not support "
            "add_total_material_balances (yet).".format(self.name)
        )

    def add_total_enthalpy_balances(
        self,
        has_heat_of_reaction=False,
        has_heat_transfer=False,
        has_work_transfer=False,
        has_enthalpy_transfer=False,
        custom_term=None,
    ):
        """
        This method constructs a set of 1D enthalpy balances indexed by time
        and phase.

        Args:
            has_heat_of_reaction: whether terms for heat of reaction should
                be included in enthalpy balance
            has_heat_transfer: whether terms for heat transfer should be
                included in enthalpy balances
            has_work_transfer: whether terms for work transfer should be
                included in enthalpy balances
            has_enthalpy_transfer: whether terms for enthalpy transfer due to
                mass transfer should be included in enthalpy balance. This
                should generally be the same as the has_mass_transfer
                argument in the material balance methods
            custom_term: a Python method which returns Pyomo expressions representing
                custom terms to be included in enthalpy balances. Method should
                accept time, length and phase list as arguments.

        Returns:
            Constraint object representing enthalpy balances
        """
        # Get dynamic and holdup flags from config block
        dynamic = self.config.dynamic
        has_holdup = self.config.has_holdup

        phase_list = self.properties.phase_list

        # Test for components that must exist prior to calling this method
        if has_heat_of_reaction:
            if not (
                hasattr(self, "rate_reaction_extent")
                or hasattr(self, "equilibrium_reaction_extent")
            ):
                raise ConfigurationError(
                    "{} extent of reaction terms must exist in order to "
                    "have heat of reaction terms. Please ensure that "
                    "add_material_balance (or equivalent) is called before"
                    " adding energy balances.".format(self.name)
                )

        # Get units from property package
        units = self.config.property_package.get_metadata().get_derived_units

        if units("energy") is not None:
            energy_l_units = units("energy") / units("length")
            power_l_units = units("power") / units("length")
        else:
            energy_l_units = None
            power_l_units = None

        # Get units for accumulation term if required
        if self.config.dynamic:
            f_time_units = self.flowsheet().time_units
            if (f_time_units is None) ^ (units("time") is None):
                raise ConfigurationError(
                    "{} incompatible time unit specification between "
                    "flowsheet and property package. Either both must use "
                    "units, or neither.".format(self.name)
                )

            if f_time_units is None:
                acc_units = None
            else:
                acc_units = energy_l_units / f_time_units

        # Create variables
        self._enthalpy_flow = Var(
            self.flowsheet().time,
            self.length_domain,
            phase_list,
            initialize=1.0,
            doc="Enthalpy flow terms",
            units=units("power"),
        )

        @self.Constraint(
            self.flowsheet().time,
            self.length_domain,
            phase_list,
            doc="Enthalpy flow linking constraints",
        )
        def enthalpy_flow_linking_constraint(b, t, x, p):
            return b._enthalpy_flow[t, x, p] == b.properties[
                t, x
            ].get_enthalpy_flow_terms(p)

        self.enthalpy_flow_dx = DerivativeVar(
            self._enthalpy_flow,
            wrt=self.length_domain,
            doc="Partial derivative of enthalpy flow wrt normalized length",
            units=units("power"),
        )

        if has_holdup:
            self.energy_holdup = Var(
                self.flowsheet().time,
                self.length_domain,
                phase_list,
                domain=Reals,
                initialize=1.0,
                doc="Enthalpy holdup per unit length",
                units=energy_l_units,
            )

        if dynamic is True:
            self.energy_accumulation = DerivativeVar(
                self.energy_holdup,
                wrt=self.flowsheet().time,
                doc="Energy accumulation per unit length",
                units=acc_units,
            )

        # Create energy balance terms as needed
        # Heat transfer term
        if has_heat_transfer:
            self.heat = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals,
                initialize=0.0,
                doc="Heat transferred per unit length",
                units=power_l_units,
            )

        # Work transfer
        if has_work_transfer:
            self.work = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals,
                initialize=0.0,
                doc="Work transferred per unit length",
                units=power_l_units,
            )

        # Enthalpy transfer
        if has_enthalpy_transfer:
            self.enthalpy_transfer = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals,
                initialize=0.0,
                doc="Enthalpy transferred due to mass transfer per unit length",
                units=power_l_units,
            )

        # Heat of Reaction
        if has_heat_of_reaction:

            @self.Expression(
                self.flowsheet().time,
                self.length_domain,
                doc="Heat of reaction term at point x",
            )
            def heat_of_reaction(b, t, x):
                if hasattr(self, "rate_reaction_extent"):
                    rate_heat = -sum(
                        b.rate_reaction_extent[t, x, r] * b.reactions[t, x].dh_rxn[r]
                        for r in self.config.reaction_package.rate_reaction_idx
                    )
                else:
                    rate_heat = 0

                if hasattr(self, "equilibrium_reaction_extent"):
                    equil_heat = -sum(
                        b.equilibrium_reaction_extent[t, x, e]
                        * b.reactions[t, x].dh_rxn[e]
                        for e in self.config.reaction_package.equilibrium_reaction_idx
                    )
                else:
                    equil_heat = 0

                return rate_heat + equil_heat

        # Energy balance equation
        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Energy balances"
        )
        def enthalpy_balances(b, t, x):
            if (
                b.config.transformation_scheme != "FORWARD"
                and x == b.length_domain.first()
            ) or (
                b.config.transformation_scheme == "FORWARD"
                and x == b.length_domain.last()
            ):
                return Constraint.Skip
            else:
                if dynamic:
                    lhs = b.length * pyunits.convert(
                        sum(b.energy_accumulation[t, x, p] for p in phase_list),
                        to_units=power_l_units,
                    )
                else:
                    lhs = 0

                rhs = b._flow_direction_term * sum(
                    b.enthalpy_flow_dx[t, x, p] for p in phase_list
                )

                if has_heat_transfer:
                    rhs += b.length * b.heat[t, x]

                if has_work_transfer:
                    rhs += b.length * b.work[t, x]

                if has_enthalpy_transfer:
                    rhs += b.length * b.enthalpy_transfer[t, x]

                if has_heat_of_reaction:
                    rhs += b.length * b.heat_of_reaction[t, x]

                if custom_term is not None:
                    rhs += b.length * custom_term(t, x)

                # TODO : Add conduction/dispersion term

                return lhs == rhs

        # Energy Holdup
        if has_holdup:
            if not hasattr(self, "phase_fraction"):
                self._add_phase_fractions()

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                phase_list,
                doc="Enthalpy holdup constraint",
            )
            def energy_holdup_calculation(b, t, x, p):
                return b.energy_holdup[t, x, p] == (
                    b._area_func(t, x)
                    * self.phase_fraction[t, x, p]
                    * b.properties[t, x].get_energy_density_terms(p)
                )

        return self.enthalpy_balances

    def add_phase_enthalpy_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_phase_enthalpy_balances.".format(self.name)
        )

    def add_phase_energy_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_phase_energy_balances.".format(self.name)
        )

    def add_total_energy_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_total_energy_balances.".format(self.name)
        )

    def add_total_pressure_balances(self, has_pressure_change=False, custom_term=None):
        """
        This method constructs a set of 1D pressure balances indexed by time.

        Args:
            has_pressure_change: whether terms for pressure change should be
                included in enthalpy balances
            custom_term: a Pyomo Expression representing custom terms to
                be included in pressure balances. Expression must be indexed
                by time and length domain
            custom_term: a Python method which returns Pyomo expressions representing
                custom terms to be included in enthalpy balances. Method should accept
                time and length as arguments.

        Returns:
            Constraint object representing pressure balances
        """

        # Get units from property package
        units = self.config.property_package.get_metadata().get_derived_units

        if units("pressure") is not None:
            pressure_l_units = units("pressure") / units("length")
        else:
            pressure_l_units = None

        # Create dP/dx terms
        self.pressure = Reference(self.properties[:, :].pressure)

        self.pressure_dx = DerivativeVar(
            self.pressure,
            wrt=self.length_domain,
            doc="Partial derivative of pressure wrt normalized length domain",
            units=units("pressure"),
        )

        # Add Momentum Balance Variables as necessary
        if has_pressure_change:
            self.deltaP = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals,
                initialize=0.0,
                doc="Pressure difference per unit length of domain",
                units=pressure_l_units,
            )

        # Momentum balance equation
        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Momentum balance"
        )
        def pressure_balance(b, t, x):
            if (
                b.config.transformation_scheme != "FORWARD"
                and x == b.length_domain.first()
            ) or (
                b.config.transformation_scheme == "FORWARD"
                and x == b.length_domain.last()
            ):
                return Constraint.Skip
            else:
                rhs = b._flow_direction_term * b.pressure_dx[t, x]

                if has_pressure_change:
                    rhs += b.length * b.deltaP[t, x]

                if custom_term is not None:
                    rhs += b.length * custom_term(t, x)

                return 0 == rhs

        return self.pressure_balance

    def add_phase_pressure_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_phase_pressure_balances.".format(self.name)
        )

    def add_phase_momentum_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_phase_momentum_balances.".format(self.name)
        )

    def add_total_momentum_balances(self, *args, **kwargs):
        """Not Supported"""
        raise BalanceTypeNotSupportedError(
            "{} OD control volumes do not support "
            "add_total_momentum_balances.".format(self.name)
        )

    def apply_transformation(self):
        """
        Method to apply DAE transformation to the Control Volume length domain.
        Transformation applied will be based on the Control Volume
        configuration arguments.
        """
        if self.length_domain.parent_block() != self:
            raise ConfigurationError(
                "{} tried to apply a DAE transformation to an external "
                "domain. To avoid complications, the apply_transformation "
                "method only supports transformation of local domains.".format(
                    self.name
                )
            )

        if self.config.finite_elements is None:
            raise ConfigurationError(
                "{} was not provided a value for the finite_elements"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )

        if (
            self.config.collocation_points is None
            and self.config.transformation_method == "dae.collocation"
        ):
            raise ConfigurationError(
                "{} was not provided a value for the collocation_points"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )

        if self.config.transformation_method == "dae.finite_difference":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                nfe=self.config.finite_elements,
                wrt=self.length_domain,
                scheme=self.config.transformation_scheme,
            )
        elif self.config.transformation_method == "dae.collocation":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                wrt=self.length_domain,
                nfe=self.config.finite_elements,
                ncp=self.config.collocation_points,
                scheme=self.config.transformation_scheme,
            )
        else:
            raise ConfigurationError(
                "{} unrecognised transformation_method, "
                "must match one of the Transformations "
                "supported by Pyomo's "
                "TransformationFactory.".format(self.name)
            )

    def estimate_states(self, always_estimate=False):
        for t in self.flowsheet().time:
            if self._flow_direction == FlowDirection.forward:
                for x in self.length_domain:
                    if not x == self.length_domain.first():
                        x_prev = self.length_domain.prev(x)
                        self._estimate_next_state(
                            self.properties[t, x_prev],
                            self.properties[t, x],
                            index=(t, x),
                            always_estimate=always_estimate,
                        )
            else:
                # Need to go backwards here
                for x in reversed(self.length_domain):
                    if not x == self.length_domain.last():
                        x_next = self.length_domain.next(x)
                        self._estimate_next_state(
                            self.properties[t, x_next],
                            self.properties[t, x],
                            index=(t, x),
                            always_estimate=always_estimate,
                        )

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                try:
                    blk.properties[t, x].model_check()
                except AttributeError:
                    _log.warning(
                        "{} ControlVolume StateBlock has no "
                        "model checks. To correct this, add a model_check"
                        " method to the associated StateBlock class.".format(blk.name)
                    )

                try:
                    blk.reactions[t, x].model_check()
                except AttributeError:
                    _log.warning(
                        "{} ControlVolume outlet reaction block has no "
                        "model check. To correct this, add a "
                        "model_check method to the associated "
                        "ReactionBlock class.".format(blk.name)
                    )

    def initialize(
        blk,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=True,
    ):
        """
        Initialization routine for 1D control volume.

        Keyword Arguments:
            state_args: a dict of arguments to be passed to the property
                package(s) to provide an initial state for initialization
                (see documentation of the specific property package) (default = {}).
            outlvl: sets output level of initialization routine
            optarg: solver options dictionary object (default=None, use
                default solver options)
            solver: str indicating which solver to use during initialization
                (default = None)
            hold_state: flag indicating whether the initialization routine
                should unfix any state variables fixed during initialization,
                (default = True). **Valid values:**
                **True** - states variables are not unfixed, and a dict of
                returned containing flags for which states were fixed
                during initialization, **False** - state variables are
                unfixed after initialization by calling the release_state
                method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization else the release state is
            triggered.
        """
        if optarg is None:
            optarg = {}

        # Get inlet state if not provided
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="control_volume")

        # Get source block
        if blk._flow_direction == FlowDirection.forward:
            source_idx = blk.length_domain.first()
        else:
            source_idx = blk.length_domain.last()
        source = blk.properties[blk.flowsheet().time.first(), source_idx]

        # Fix source state and get state_args if not provided
        source_flags = {}
        state_dict = source.define_port_members()

        for k in state_dict.keys():
            if state_dict[k].is_indexed():
                source_flags[k] = {}
                for m in state_dict[k].keys():
                    source_flags[k][m] = state_dict[k][m].fixed
                    state_dict[k][m].fix()
            else:
                source_flags[k] = state_dict[k].fixed
                state_dict[k].fix()

        if state_args is None:
            # If no initial guesses provided, estimate values for states
            blk.estimate_states(always_estimate=True)

        # Initialize state blocks
        flags = blk.properties.initialize(
            state_args=state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        try:
            # TODO: setting state_vars_fixed may not work for heterogeneous
            # systems where a second control volume is involved, as we cannot
            # assume those state vars are also fixed. For now, heterogeneous
            # reactions should ignore the state_vars_fixed argument and always
            # check their state_vars.
            blk.reactions.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_vars_fixed=True,
            )
        except AttributeError:
            pass

        init_log.info("Initialization Complete")

        # Unfix state variables except for source block
        blk.properties.release_state(flags)

        if hold_state is True:
            return source_flags
        else:
            blk.release_state(source_flags)

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags: dict containing information of which state variables
                were fixed during initialization, and should now be
                unfixed. This dict is returned by initialize if
                hold_state = True.
            outlvl: sets output level of logging

        Returns:
            None
        """
        # Get source block
        if blk._flow_direction == FlowDirection.forward:
            source_idx = blk.length_domain.first()
        else:
            source_idx = blk.length_domain.last()
        source = blk.properties[blk.flowsheet().time.first(), source_idx]

        # Set fixed attribute on state vars based on flags
        state_dict = source.define_port_members()

        for k in state_dict.keys():
            if state_dict[k].is_indexed():
                for m in state_dict[k].keys():
                    state_dict[k][m].fixed = flags[k][m]
            else:
                state_dict[k].fixed = flags[k]

    def _add_phase_fractions(self):
        """
        This method constructs the phase_fraction variables for the control
        volume, and the associated constraint on the sum of phase_fractions
        == 1. For systems with only one phase, phase_fraction is created as a
        Pyomo Expression with a value of 1.

        Args:
            None

        Returns:
            None
        """
        phase_list = self.properties.phase_list

        if len(phase_list) > 1:
            self.phase_fraction = Var(
                self.flowsheet().time,
                self.length_domain,
                phase_list,
                initialize=1 / len(phase_list),
                doc="Volume fraction of holdup by phase",
            )

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Sum of phase fractions == 1",
            )
            def sum_of_phase_fractions(b, t, x):
                return 1 == sum(b.phase_fraction[t, x, p] for p in phase_list)

        else:

            @self.Expression(
                self.flowsheet().time,
                self.length_domain,
                phase_list,
                doc="Volume fraction of holdup by phase",
            )
            def phase_fraction(self, t, x, p):
                return 1

    def _area_func(b, t, x):
        if b.config.area_definition == DistributedVars.uniform:
            return b.area
        return b.area[t, x]

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):
        """
        No report method defined for ControlVolume1D class. This is due to the
        difficulty of presenting spatially discretized data in a readable form
        without plotting.
        """
        raise NotImplementedError(
            """
                Due to the difficultly in presenting spatially distributed data
                in a clean format, ControlVolume1D does not currently support
                the report method."""
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        phase_list = self.properties.phase_list
        phase_component_set = self.properties.phase_component_set

        # Default scale factors
        # If the parent component of an indexed component has a scale factor,
        # but some of the data objects don't, propagate the indexed component
        # scale factor to the missing scaling factors.
        iscale.propagate_indexed_component_scaling_factors(self)

        # Set scaling for geometry variables
        if hasattr(self, "area"):
            for v in self.area.values():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(self.area, default=1, warning=True)
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "length"):
            if iscale.get_scaling_factor(self.length) is None:
                iscale.set_scaling_factor(self.length, 1)

        if hasattr(self, "phase_fraction"):
            for v in self.phase_fraction.values():
                if iscale.get_scaling_factor(v) is None:
                    # phase fraction typically between 0.1 and 1
                    iscale.set_scaling_factor(v, 10)

        # Set scaling factors for common material balance variables
        if hasattr(self, "material_holdup"):
            for (t, x, p, j), v in self.material_holdup.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(self._area_func(t, x))
                    sf *= iscale.get_scaling_factor(self.phase_fraction[t, x, p])
                    sf *= iscale.get_scaling_factor(
                        self.properties[t, x].get_material_density_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "material_accumulation"):
            for (t, x, p, j), v in self.material_accumulation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].get_material_flow_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "_flow_terms"):
            for (t, x, p, j), v in self._flow_terms.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].get_material_flow_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "material_flow_dx"):
            for (t, x, p, j), v in self.material_flow_dx.items():
                if iscale.get_scaling_factor(v) is None:
                    # As domain is normalized, derivative should have same
                    # scale as flow
                    sf = iscale.get_scaling_factor(self._flow_terms[t, x, p, j])
                    iscale.set_scaling_factor(v, sf)

        # Control Volume has no way of knowing how best to scale
        # reaction extents - this is something only the unit model can provide
        # This also applies to the phase_equilibrium_generation term.

        if hasattr(self, "rate_reaction_generation"):
            for (t, x, p, j), v in self.rate_reaction_generation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(self.rate_reaction_extent[t, x, :])
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "equilibrium_reaction_generation"):
            for (t, x, p, j), v in self.equilibrium_reaction_generation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(
                        self.equilibrium_reaction_extent[t, x, ...]
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "inherent_reaction_generation"):
            for (t, x, p, j), v in self.inherent_reaction_generation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(
                        self.inherent_reaction_extent[t, x, ...]
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "mass_transfer_term"):
            for (t, x, p, j), v in self.mass_transfer_term.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].get_material_flow_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        # Set scaling factors for element balance variables
        if hasattr(self, "elemental_flow_term"):
            for (t, x, e), v in self.elemental_flow_term.items():
                flow_basis = self.properties[t, x].get_material_flow_basis()

                sf = iscale.min_scaling_factor(
                    [
                        self.properties[t, x].get_material_density_terms(p, j)
                        for (p, j) in phase_component_set
                    ],
                    default=1,
                    warning=True,
                )
                if flow_basis == MaterialFlowBasis.molar:
                    sf *= 1
                elif flow_basis == MaterialFlowBasis.mass:
                    # MW scaling factor is the inverse of its value
                    sf *= value(self.properties[t, x].mw_comp[j])

                iscale.set_scaling_factor(v, sf)

        if hasattr(self, "elemental_flow_dx"):
            for (t, x, e), v in self.elemental_flow_dx.items():
                if iscale.get_scaling_factor(v) is None:
                    # As domain is normalized, scale should be equal to flow
                    sf = iscale.get_scaling_factor(self.elemental_flow_term[t, x, e])
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "element_holdup"):
            for (t, x, e), v in self.element_holdup.items():
                flow_basis = self.properties[t, x].get_material_flow_basis()
                sf_list = []
                for p, j in phase_component_set:
                    if flow_basis == MaterialFlowBasis.molar:
                        sf = 1
                    elif flow_basis == MaterialFlowBasis.mass:
                        # MW scaling factor is the inverse of its value
                        sf = value(self.properties[t, x].mw_comp[j])
                    sf *= iscale.get_scaling_factor(self.phase_fraction[t, x, p])
                    sf *= iscale.get_scaling_factor(
                        self.properties[t, x].get_material_density_terms(p, j),
                        default=1,
                        warning=True,
                    )
                    sf *= value(self.properties[t, x].params.element_comp[j][e]) ** -1
                    sf_list.append(sf)
                sf_h = min(sf_list) * iscale.get_scaling_factor(self._area_func(t, x))
                iscale.set_scaling_factor(v, sf_h)

        if hasattr(self, "element_accumulation"):
            for (t, x, e), v in self.element_accumulation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(
                        self.elemental_flow_term[t, x, ...], default=1, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "elemental_mass_transfer_term"):
            for (t, x, e), v in self.elemental_mass_transfer_term.items():
                # minimum scaling factor for elemental_flow terms
                sf_list = []
                flow_basis = self.properties[t, x].get_material_flow_basis()
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(
                        self.elemental_flow_term[t, x, ...], default=1, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        # Set scaling factors for enthalpy balance variables
        if hasattr(self, "_enthalpy_flow"):
            for (t, x, p), v in self._enthalpy_flow.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].get_enthalpy_flow_terms(p),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "enthalpy_flow_dx"):
            for (t, x, p), v in self.enthalpy_flow_dx.items():
                if iscale.get_scaling_factor(v) is None:
                    # Normalized domain, so scale should be the same as flow
                    sf = iscale.get_scaling_factor(self._enthalpy_flow[t, x, p])
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "energy_holdup"):
            for (t, x, p), v in self.energy_holdup.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(self._area_func(t, x))
                    sf = iscale.get_scaling_factor(self.phase_fraction[t, x, p])
                    sf *= iscale.get_scaling_factor(
                        self.properties[t, x].get_energy_density_terms(p),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "energy_accumulation"):
            for (t, x, p), v in self.energy_accumulation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].get_enthalpy_flow_terms(p),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "heat"):
            for v in self.heat.values():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.heat, default=1e-6, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "work"):
            for v in self.work.values():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.work, default=1e-6, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "enthalpy_transfer"):
            for (t, x), v in self.enthalpy_transfer.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.min_scaling_factor(
                        [
                            self.properties[t, x].get_enthalpy_flow_terms(p)
                            for p in phase_list
                        ]
                    )
                    iscale.set_scaling_factor(v, sf)

        # Set scaling for momentum balance variables
        if hasattr(self, "pressure_dx"):
            for (t, x), v in self.pressure_dx.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.properties[t, x].pressure, default=1, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        if hasattr(self, "deltaP"):
            for (t, x), v in self.deltaP.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = 10 * iscale.get_scaling_factor(
                        self.properties[t, x].pressure, default=1, warning=True
                    )
                    iscale.set_scaling_factor(v, sf)

        # Transform constraints in order of appearance
        if hasattr(self, "material_flow_linking_constraints"):
            for (t, x, p, j), c in self.material_flow_linking_constraints.items():
                sf = iscale.get_scaling_factor(self._flow_terms[t, x, p, j])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "material_holdup_calculation"):
            for (t, x, p, j), c in self.material_holdup_calculation.items():
                sf = iscale.get_scaling_factor(self.material_holdup[t, x, p, j])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "rate_reaction_stoichiometry_constraint"):
            for (t, x, p, j), c in self.rate_reaction_stoichiometry_constraint.items():
                sf = iscale.get_scaling_factor(
                    self.rate_reaction_generation[t, x, p, j]
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "equilibrium_reaction_stoichiometry_constraint"):
            for (
                t,
                x,
                p,
                j,
            ), c in self.equilibrium_reaction_stoichiometry_constraint.items():
                sf = iscale.get_scaling_factor(
                    self.equilibrium_reaction_generation[t, x, p, j]
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "material_balances"):
            mb_type = self._constructed_material_balance_type
            if mb_type == MaterialBalanceType.componentPhase:
                for (t, x, p, j), c in self.material_balances.items():
                    sf = iscale.get_scaling_factor(self._flow_terms[t, x, p, j])
                    iscale.constraint_scaling_transform(c, sf, overwrite=False)
            elif mb_type == MaterialBalanceType.componentTotal:
                for (t, x, j), c in self.material_balances.items():
                    sf = iscale.min_scaling_factor(
                        [
                            self._flow_terms[t, x, p, j]
                            for p in phase_list
                            if (p, j) in phase_component_set
                        ]
                    )
                    iscale.constraint_scaling_transform(c, sf, overwrite=False)
            else:
                _log.warning(f"Unknown material balance type {mb_type}")

        if hasattr(self, "elemental_flow_constraint"):
            for (t, x, e), c in self.elemental_flow_constraint.items():
                sf = iscale.get_scaling_factor(self.elemental_flow_term[t, x, e])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "element_balances"):
            for (t, x, e), c in self.element_balances.items():
                sf = iscale.get_scaling_factor(self.elemental_flow_dx[t, x, e])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "elemental_holdup_calculation"):
            for (t, x, e), c in self.elemental_holdup_calculation.items():
                sf = iscale.get_scaling_factor(self.element_holdup[t, x, e])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "enthalpy_flow_linking_constraint"):
            for (t, x, p), c in self.enthalpy_flow_linking_constraint.items():
                sf = iscale.get_scaling_factor(self._enthalpy_flow[t, x, p])
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "enthalpy_balances"):
            for (t, x), c in self.enthalpy_balances.items():
                sf = iscale.min_scaling_factor(
                    [self._enthalpy_flow[t, x, p] for p in phase_list]
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "energy_holdup_calculation"):
            for (t, x, p), c in self.energy_holdup_calculation.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.energy_holdup[t, x, p]),
                    overwrite=False,
                )

        if hasattr(self, "pressure_balance"):
            for (t, x), c in self.pressure_balance.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.properties[t, x].pressure, default=1e-5
                    ),
                    overwrite=False,
                )

        if hasattr(self, "sum_of_phase_fractions"):
            for (t, x), c in self.sum_of_phase_fractions.items():
                sf = iscale.min_scaling_factor(
                    [self.phase_fraction[t, x, p] for p in phase_list]
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        if hasattr(self, "material_flow_dx_disc_eq"):
            for (t, x, p, j), c in self.material_flow_dx_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.material_flow_dx[t, x, p, j]),
                    overwrite=False,
                )

        if hasattr(self, "material_accumulation_disc_eq"):
            for (t, x, p, j), c in self.material_accumulation_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.material_accumulation[t, x, p, j]),
                    overwrite=False,
                )

        # Scaling for discretization equations
        if hasattr(self, "enthalpy_flow_dx_disc_eq"):
            for (t, x, p), c in self.enthalpy_flow_dx_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.enthalpy_flow_dx[t, x, p]),
                    overwrite=False,
                )

        if hasattr(self, "energy_accumulation_disc_eq"):
            for (t, x, p), c in self.energy_accumulation_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.energy_accumulation[t, x, p]),
                    overwrite=False,
                )

        if hasattr(self, "pressure_dx_disc_eq"):
            for (t, x), c in self.pressure_dx_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.pressure_dx[t, x]),
                    overwrite=False,
                )

        if hasattr(self, "elemental_flow_dx_disc_eq"):
            for (t, x, e), c in self.elemental_flow_dx_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.elemental_flow_dx[t, x, e]),
                    overwrite=False,
                )

        if hasattr(self, "element_accumulation_disc_eq"):
            for (t, x, e), c in self.element_accumulation_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.element_accumulation[t, x, e]),
                    overwrite=False,
                )



# coding: utf-8
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""__init__.py for idaes module

Set up logging for the idaes module, and import plugins.
"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

import os
import copy
import logging
from typing import Optional, List

from pyomo.common.fileutils import find_library

from . import config
from .ver import __version__  # noqa


def _handle_optional_compat_activation(
    env_var: str = "IDAES_ACTIVATE_V1_COMPAT",
):
    _log = logging.getLogger("idaes_v1_compat")
    found_in_env = os.environ.get(env_var, None)
    if found_in_env:
        _log.warning(
            "Found environment variable %s=%s. Activating IDAES V1 compatibility.",
            env_var,
            found_in_env,
        )
        try:
            # Only need to import this if required
            # pylint: disable=import-outside-toplevel
            from _idaes_v1_compat import activate
        except ImportError:
            _log.error("Required package _idaes_v1_compat not found")
        else:
            activate()


_handle_optional_compat_activation()

_log = logging.getLogger(__name__)

# Standard locations for config file, binary libraries and executables, ...
data_directory, bin_directory, testing_directory = config.get_data_directory()
# To avoid a circular import the config module doesn't import idaes, but
# some functions in the config module that are executed later use this
# these directories are static from here on.
config.data_directory = data_directory
config.bin_directory = bin_directory
config.testing_directory = testing_directory

# Set the path for the global and local config files
if data_directory is not None:
    _global_config_file = os.path.join(data_directory, "idaes.conf")
else:
    _global_config_file = None
_local_config_file = "idaes.conf"

# Create the general IDAES configuration block, with default config
cfg = config._new_idaes_config_block()  # pylint: disable=protected-access
config.reconfig(cfg)
# read global config and overwrite provided config options
config.read_config(_global_config_file, cfg=cfg)
# read local config and overwrite provided config options
config.read_config(_local_config_file, cfg=cfg)

# Setup the environment so solver executables can be run
config.setup_environment(bin_directory, cfg.use_idaes_solvers)

# Debug log for basic testing of the logging config
_log.debug("'idaes' logger debug test")


# TODO: Remove once AMPL bug is fixed
# TODO: https://github.com/ampl/asl/issues/13
# There appears to be a bug in the ASL which causes terminal failures
# if you try to create multiple ASL structs with different external
# functions in the same process. This causes pytest to crash during testing.
# To avoid this, register all known external functions at initialization.
def _ensure_external_functions_libs_in_env(
    ext_funcs: List[str], var_name: str = "AMPLFUNC", sep: str = "\n"
):
    libraries_str = os.environ.get(var_name, "")
    libraries = [lib for lib in libraries_str.split(sep) if lib.strip()]
    for func_name in ext_funcs:
        lib: Optional[str] = find_library(os.path.join(bin_directory, func_name))
        if lib is not None and lib not in libraries:
            libraries.append(lib)
    os.environ[var_name] = sep.join(libraries)


_ensure_external_functions_libs_in_env(
    ["cubic_roots", "general_helmholtz_external", "functions"]
)


def _create_data_dir():
    """Create the IDAES directory to store data files in."""
    config.create_dir(data_directory)


def _create_bin_dir(bd=None):
    """Create the IDAES directory to store executable files in.

    Args:
        bd: alternate binary directory, used for testing
    """
    _create_data_dir()
    if bd is None:
        bd = bin_directory
    config.create_dir(bd)


def _create_testing_dir():
    """Create an idaes testing directory"""
    _create_data_dir()
    config.create_dir(testing_directory)


if data_directory is not None:
    try:
        _create_data_dir()
    except FileNotFoundError:
        pass  # the standard place for this doesn't exist, shouldn't be a show stopper

    try:
        _create_bin_dir()
    except FileNotFoundError:
        pass  # the standard place for this doesn't exist, shouldn't be a show stopper

    try:
        _create_testing_dir()
    except FileNotFoundError:
        pass  # the standard place for this doesn't exist, shouldn't be a show stopper


def reconfig():
    return config.reconfig(cfg)


def read_config(val):
    return config.read_config(val=val, cfg=cfg)


def write_config(path, default=False):
    _cfg = None if default else cfg
    return config.write_config(path=path, cfg=_cfg)


class temporary_config_ctx(object):
    def __enter__(self):
        self.orig_config = copy.deepcopy(cfg)

    def __exit__(self, exc_type, exc_value, traceback):
        global cfg  # pylint: disable=global-statement
        cfg = self.orig_config
        reconfig()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for saponification property package example.
Authors: Andrew Lee
"""
from math import exp
import pytest

from pyomo.environ import ConcreteModel, Constraint, Param, Suffix, units, value, Var

from idaes.core import MaterialFlowBasis
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
    ReactionBlock,
    SaponificationReactionScaler,
)
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.core.solvers import get_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
            property_package=model.pparams
        )

        return model

    @pytest.mark.unit
    def test_config(self, model):
        assert len(model.rparams.config) == 2

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rparams.reaction_block_class is ReactionBlock

        assert len(model.rparams.rate_reaction_idx) == 1
        for i in model.rparams.rate_reaction_idx:
            assert i == "R1"

        assert len(model.rparams.rate_reaction_stoichiometry) == 5
        for i in model.rparams.rate_reaction_stoichiometry:
            assert i in [
                ("R1", "Liq", "NaOH"),
                ("R1", "Liq", "EthylAcetate"),
                ("R1", "Liq", "SodiumAcetate"),
                ("R1", "Liq", "Ethanol"),
                ("R1", "Liq", "H2O"),
            ]

        assert isinstance(model.rparams.arrhenius, Param)
        assert value(model.rparams.arrhenius) == 3.132e6

        assert isinstance(model.rparams.energy_activation, Param)
        assert value(model.rparams.energy_activation) == 43000

        assert isinstance(model.rparams.dh_rxn, Param)
        assert len(model.rparams.dh_rxn) == 1
        for i in model.rparams.dh_rxn:
            assert value(model.rparams.dh_rxn[i]) == -49000


class TestReactionBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
            property_package=model.pparams
        )

        model.props = model.pparams.build_state_block([1])

        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert model.rxns[1].conc_mol_comp_ref is model.props[1].conc_mol_comp
        assert model.rxns[1].temperature_ref is model.props[1].temperature
        assert model.rxns[1].dh_rxn is model.rparams.dh_rxn

        assert model.rxns[1].default_scaler is SaponificationReactionScaler

    @pytest.mark.unit
    def test_rate_constant(self, model):
        assert isinstance(model.rxns[1].k_rxn, Var)
        assert isinstance(model.rxns[1].arrhenius_eqn, Constraint)

    @pytest.mark.unit
    def test_rxn_rate(self, model):
        assert isinstance(model.rxns[1].reaction_rate, Var)
        assert isinstance(model.rxns[1].rate_expression, Constraint)

    @pytest.mark.unit
    def test_get_reaction_rate_basis(self, model):
        assert model.rxns[1].get_reaction_rate_basis() == MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_model_check(self, model):
        assert model.rxns[1].model_check() is None

    @pytest.mark.unit
    def test_initialize(self, model):
        assert model.rxns.initialize(outlvl=1) is None

    def check_units(self, model):
        units.assert_units_consistent(model)


class TestSaponificationReactionScaler(object):
    @pytest.mark.unit
    def test_variable_scaling_routine(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
            property_package=model.pparams
        )

        model.props = model.pparams.build_state_block([1])
        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        # Trigger build of reaction properties
        model.rxns[1].reaction_rate

        scaler = model.rxns[1].default_scaler()
        assert isinstance(scaler, SaponificationReactionScaler)

        scaler.variable_scaling_routine(model.rxns[1])

        assert isinstance(model.rxns[1].scaling_factor, Suffix)

        sfx = model.rxns[1].scaling_factor
        assert len(sfx) == 2
        assert sfx[model.rxns[1].k_rxn] == pytest.approx(
            1 / (3.132e6 * exp(-43000 / (8.31446262 * 298.15))), rel=1e-8
        )
        assert sfx[model.rxns[1].reaction_rate["R1"]] == pytest.approx(1e2, rel=1e-8)

    @pytest.mark.unit
    def test_constraint_scaling_routine(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
            property_package=model.pparams
        )

        model.props = model.pparams.build_state_block([1])
        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        # Trigger build of reaction properties
        model.rxns[1].reaction_rate

        scaler = model.rxns[1].default_scaler()
        assert isinstance(scaler, SaponificationReactionScaler)

        scaler.constraint_scaling_routine(model.rxns[1])

        assert isinstance(model.rxns[1].scaling_factor, Suffix)

        sfx = model.rxns[1].scaling_factor
        assert len(sfx) == 2
        assert sfx[model.rxns[1].arrhenius_eqn] == pytest.approx(1, rel=1e-8)
        assert sfx[model.rxns[1].rate_expression["R1"]] == pytest.approx(1e-4, rel=1e-8)

    @pytest.mark.unit
    def test_scale_model(self):
        model = ConcreteModel()
        model.pparams = SaponificationParameterBlock()
        model.rparams = SaponificationReactionParameterBlock(
            property_package=model.pparams
        )

        model.props = model.pparams.build_state_block([1])
        model.rxns = model.rparams.build_reaction_block([1], state_block=model.props)

        # Trigger build of reaction properties
        model.rxns[1].reaction_rate

        scaler = model.rxns[1].default_scaler()
        assert isinstance(scaler, SaponificationReactionScaler)

        scaler.scale_model(model.rxns[1])

        assert isinstance(model.rxns[1].scaling_factor, Suffix)

        sfx = model.rxns[1].scaling_factor
        k_rxn_sf = 1 / (3.132e6 * exp(-43000 / (8.31446262 * 298.15)))
        assert len(sfx) == 4
        assert sfx[model.rxns[1].k_rxn] == pytest.approx(k_rxn_sf, rel=1e-8)
        assert sfx[model.rxns[1].reaction_rate["R1"]] == pytest.approx(1e2, rel=1e-8)
        assert sfx[model.rxns[1].arrhenius_eqn] == pytest.approx(k_rxn_sf, rel=1e-8)
        assert sfx[model.rxns[1].rate_expression["R1"]] == pytest.approx(
            1e-4 * k_rxn_sf, rel=1e-8
        )



# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains common constants of use in process systems engineering.

All units are SI, and generally expressed to 9 significant figures where
available
"""

__author__ = "Andrew Lee"

import math

from pyomo.environ import units


class Constants:
    """
    IDAES Constants class
    """

    # -------------------------------------------------------------------------
    # General geometric relationships
    pi = math.pi

    # -------------------------------------------------------------------------
    # Constants used as fundamental definitions in SI, from
    # https://www.bipm.org/utils/common/pdf/si-brochure/SI-Brochure-9.pdf
    avogadro_number = 6.02214076e23 / units.mol

    boltzmann_constant = 1.38064900e-23 * units.joule / units.degK

    elemental_charge = 1.602176634e-19 * units.coulomb

    planck_constant = 6.62607015e-34 * units.joule * units.second

    speed_light = 299792458 * units.m / units.s  # in a vacuum

    # -------------------------------------------------------------------------
    # Constants derived from fundamental constants

    # Faraday constant = elemental charge * Avogadro's constant
    faraday_constant = 96485.33212 * units.coulomb / units.mol

    # Gas constant = Avogadro's constant * Boltzmann's constant
    gas_constant = 8.314462618 * units.joule / units.mol / units.degK

    # Stefan-Boltzmann constant
    # Function of Boltzmann constant, pi and speed of light
    stefan_constant = 5.67037442e-8 * units.watt / units.metre**2 / units.degK**4

    # -------------------------------------------------------------------------
    # Other constants - all values sourced from NIST to available uncertainty
    # All values retrieved 8th Jan 2020 unless otherwise noted

    # https://physics.nist.gov/cgi-bin/cuu/Value?gn
    acceleration_gravity = 9.80665 * units.metre / units.second**2

    # https://physics.nist.gov/cgi-bin/cuu/Value?bg
    gravitational_constant = 6.67430e-11 * units.metre**3 / units.kg / units.second**2

    # https://physics.nist.gov/cgi-bin/cuu/Value?me
    mass_electron = 9.1093837015e-31 * units.kilogram

    # https://physics.nist.gov/cgi-bin/cuu/Value?ep0
    # 8th April 2021
    vacuum_electric_permittivity = 8.8541878128e-12 * units.farad * units.metre**-1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for eNRTL methods

Author: Andrew Lee
"""
import pytest

from pyomo.environ import (
    ConcreteModel,
    Expression,
    exp,
    log,
    Set,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_equivalent

from idaes.core import AqueousPhase, Solvent, Solute, Apparent, Anion, Cation
from idaes.core.util.constants import Constants
from idaes.models.properties.modular_properties.eos.enrtl import ENRTL
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
    StateIndex,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.pure.electrolyte import (
    relative_permittivity_constant,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


def dummy_method(b, *args, **kwargs):
    return 42.0 * pyunits.mol / pyunits.m**3


configuration = {
    "components": {
        "H2O": {
            "type": Solvent,
            "dens_mol_liq_comp": dummy_method,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (18e-3, pyunits.kg / pyunits.mol),
                "relative_permittivity_liq_comp": 101,
            },
        },
        "C6H12": {
            "type": Solute,
            "dens_mol_liq_comp": dummy_method,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (84e-3, pyunits.kg / pyunits.mol),
                "relative_permittivity_liq_comp": 102,
            },
        },
        "NaCl": {"type": Apparent, "dissociation_species": {"Na+": 1, "Cl-": 1}},
        "HCl": {"type": Apparent, "dissociation_species": {"H+": 1, "Cl-": 1}},
        "NaOH": {"type": Apparent, "dissociation_species": {"Na+": 1, "OH-": 1}},
        "Na+": {"type": Cation, "charge": +1},
        "H+": {"type": Cation, "charge": +1},
        "Cl-": {"type": Anion, "charge": -1},
        "OH-": {"type": Anion, "charge": -1},
    },
    "phases": {"Liq": {"type": AqueousPhase, "equation_of_state": ENRTL}},
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 300,
}


class TestParameters(object):
    @pytest.mark.unit
    def test_parameters_no_assignment(self):
        m = ConcreteModel()

        m.params = GenericParameterBlock(**configuration)

        assert isinstance(m.params.Liq.ion_pair_set, Set)
        assert len(m.params.Liq.ion_pair_set) == 4
        for p in m.params.Liq.ion_pair_set:
            assert p in [("Na+, Cl-"), ("Na+, OH-"), ("H+, Cl-"), ("H+, OH-")]

        assert isinstance(m.params.Liq.component_pair_set, Set)
        assert len(m.params.Liq.component_pair_set) == 32
        assert isinstance(m.params.Liq.component_pair_set_symmetric, Set)
        assert len(m.params.Liq.component_pair_set_symmetric) == 17

        assert isinstance(m.params.Liq.alpha, Var)
        assert len(m.params.Liq.alpha) == 17
        for i, j in m.params.Liq.alpha:
            if i != j:
                assert (j, i) not in m.params.Liq.alpha
            if (i, j) in [("C6H12", "C6H12"), ("H2O", "H2O"), ("H2O", "C6H12")]:
                assert m.params.Liq.alpha[(i, j)].value == 0.3
                assert m.params.Liq.alpha[(i, j)].fixed
            else:
                assert m.params.Liq.alpha[(i, j)].value == 0.2
                assert m.params.Liq.alpha[(i, j)].fixed

        assert isinstance(m.params.Liq.tau, Var)
        assert len(m.params.Liq.tau) == 32
        for i, j in m.params.Liq.tau:
            assert m.params.Liq.tau[(i, j)].value == 0
            assert m.params.Liq.tau[(i, j)].fixed

    @pytest.mark.unit
    def test_parameters_assignment(self):
        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_alpha"] = {}
        test_config["parameter_data"]["Liq_alpha"][("H2O", "Na+, Cl-")] = 0.6
        test_config["parameter_data"]["Liq_tau"] = {}
        test_config["parameter_data"]["Liq_tau"][("H2O", "Na+, Cl-")] = 0.1

        m = ConcreteModel()

        m.params = GenericParameterBlock(**test_config)

        assert isinstance(m.params.Liq.alpha, Var)
        assert len(m.params.Liq.alpha) == 17
        for i, j in m.params.Liq.alpha:
            if i != j:
                assert (j, i) not in m.params.Liq.alpha
            if (i, j) == ("H2O", "Na+, Cl-"):
                assert m.params.Liq.alpha[(i, j)].value == 0.6
                assert m.params.Liq.alpha[(i, j)].fixed
            elif (i, j) in [("C6H12", "C6H12"), ("H2O", "H2O"), ("H2O", "C6H12")]:
                assert m.params.Liq.alpha[(i, j)].value == 0.3
                assert m.params.Liq.alpha[(i, j)].fixed
            else:
                assert m.params.Liq.alpha[(i, j)].value == 0.2
                assert m.params.Liq.alpha[(i, j)].fixed

        assert isinstance(m.params.Liq.tau, Var)
        assert len(m.params.Liq.tau) == 32
        for i, j in m.params.Liq.tau:
            print(i, j)
            if (i, j) == ("H2O", "Na+, Cl-"):
                assert m.params.Liq.tau[(i, j)].value == 0.1
                assert m.params.Liq.tau[(i, j)].fixed
            else:
                assert m.params.Liq.tau[(i, j)].value == 0
                assert m.params.Liq.tau[(i, j)].fixed

    @pytest.mark.unit
    def test_parameters_unsymmetric_alpha(self):
        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_alpha"] = {}
        test_config["parameter_data"]["Liq_alpha"][("H2O", "Na+, Cl-")] = 0.6
        test_config["parameter_data"]["Liq_alpha"][("Na+, Cl-", "H2O")] = 0.8

        m = ConcreteModel()

        # TODO: Having trouble getting regex to match component tuple
        # Using a wildcard for now
        with pytest.raises(
            ConfigurationError,
            match="params.Liq eNRTL alpha parameter assigned "
            "non-symmetric value for pair (.+?). Please assign "
            "only one value for component pair.",
        ):
            m.params = GenericParameterBlock(**test_config)

    @pytest.mark.unit
    def test_parameters_alpha_symmetry_duplicate(self, caplog):
        caplog.set_level(
            idaeslog.INFO,
            logger=(
                "idaes.models.properties.modular_properties." "eos.enrtl_parameters"
            ),
        )

        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_alpha"] = {}
        test_config["parameter_data"]["Liq_alpha"][("H2O", "Na+, Cl-")] = 0.6
        test_config["parameter_data"]["Liq_alpha"][("Na+, Cl-", "H2O")] = 0.6

        m = ConcreteModel()

        m.params = GenericParameterBlock(**test_config)

        assert (
            "eNRTL alpha value provided for both ('H2O', 'Na+, Cl-') and "
            "('Na+, Cl-', 'H2O'). It is only necessary to provide a "
            "value for one of these due to symmetry." in caplog.text
        )

    @pytest.mark.unit
    def test_parameters_alpha_unused_parameter(self):
        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_alpha"] = {}
        test_config["parameter_data"]["Liq_alpha"][("H2O", "Na+")] = 0.6

        m = ConcreteModel()

        # TODO: Having trouble getting regex to match component tuple
        # Using a wildcard for now
        with pytest.raises(
            ConfigurationError,
            match="params.Liq eNRTL alpha parameter provided "
            "for invalid component pair (.+?). Please check "
            "typing and only provide parameters for valid "
            "species pairs.",
        ):
            m.params = GenericParameterBlock(**test_config)

    @pytest.mark.unit
    def test_parameters_tau_asymmetric(self):
        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_tau"] = {}
        test_config["parameter_data"]["Liq_tau"][("H2O", "Na+, Cl-")] = 0.1
        test_config["parameter_data"]["Liq_tau"][("Na+, Cl-", "H2O")] = -0.1

        m = ConcreteModel()

        m.params = GenericParameterBlock(**test_config)

        assert isinstance(m.params.Liq.tau, Var)
        assert len(m.params.Liq.tau) == 32
        for i, j in m.params.Liq.tau:
            print(i, j)
            if (i, j) == ("H2O", "Na+, Cl-"):
                assert m.params.Liq.tau[(i, j)].value == 0.1
                assert m.params.Liq.tau[(i, j)].fixed
            elif (i, j) == ("Na+, Cl-", "H2O"):
                assert m.params.Liq.tau[(i, j)].value == -0.1
                assert m.params.Liq.tau[(i, j)].fixed
            else:
                assert m.params.Liq.tau[(i, j)].value == 0
                assert m.params.Liq.tau[(i, j)].fixed

    @pytest.mark.unit
    def test_parameters_tau_unused_parameter(self):
        test_config = dict(configuration)
        test_config["parameter_data"] = {}
        test_config["parameter_data"]["Liq_tau"] = {}
        test_config["parameter_data"]["Liq_tau"][("H2O", "Na+")] = 0.6

        m = ConcreteModel()

        # TODO: Having trouble getting regex to match component tuple
        # Using a wildcard for now
        with pytest.raises(
            ConfigurationError,
            match="params.Liq eNRTL tau parameter provided for "
            "invalid component pair (.+?). Please check typing "
            "and only provide parameters for valid species "
            "pairs.",
        ):
            m.params = GenericParameterBlock(**test_config)


class TestStateBlockSymmetric(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.params = GenericParameterBlock(**configuration)

        m.state = m.params.build_state_block([1])

        # Need to set a value of T for checking expressions later
        m.state[1].temperature.set_value(300)

        return m

    @pytest.mark.unit
    def test_common(self, model):
        # Reference state composition
        assert isinstance(model.state[1].Liq_x_ref, Expression)
        assert len(model.state[1].Liq_x_ref) == 6
        for k in model.state[1].Liq_x_ref:
            assert k in ["H2O", "C6H12", "Na+", "H+", "Cl-", "OH-"]
            if k in ["H2O", "C6H12"]:
                assert str(model.state[1].Liq_x_ref[k].expr) == str(0.0)
            else:
                assert str(model.state[1].Liq_x_ref[k].expr) == str(
                    model.state[1].mole_frac_phase_comp_true["Liq", k]
                    / (
                        model.state[1].mole_frac_phase_comp_true["Liq", "Cl-"]
                        + model.state[1].mole_frac_phase_comp_true["Liq", "OH-"]
                        + model.state[1].mole_frac_phase_comp_true["Liq", "Na+"]
                        + model.state[1].mole_frac_phase_comp_true["Liq", "H+"]
                    )
                )

        assert isinstance(model.state[1].Liq_X, Expression)
        assert len(model.state[1].Liq_X) == 6
        for j in model.state[1].Liq_X:
            if j in ["H2O", "C6H12"]:
                # _X should be mole_frac_phase_comp_true
                assert str(model.state[1].Liq_X[j].expr) == str(
                    model.state[1].mole_frac_phase_comp_true["Liq", j]
                )
            else:
                # _X should be mutiplied by |charge|
                assert str(model.state[1].Liq_X[j].expr) == str(
                    model.state[1].mole_frac_phase_comp_true["Liq", j]
                    * abs(model.params.get_component(j).config.charge)
                )

        assert isinstance(model.state[1].Liq_X_ref, Expression)
        assert len(model.state[1].Liq_X_ref) == 6
        for j in model.state[1].Liq_X_ref:
            if j in ["H2O", "C6H12"]:
                # _X should be mole_frac_phase_comp_true
                assert str(model.state[1].Liq_X_ref[j].expr) == str(
                    model.state[1].Liq_x_ref[j]
                )
            else:
                # _X should be mutiplied by |charge|
                assert str(model.state[1].Liq_X_ref[j].expr) == str(
                    model.state[1].Liq_x_ref[j]
                    * abs(model.params.get_component(j).config.charge)
                )

        assert isinstance(model.state[1].Liq_Y, Expression)
        assert len(model.state[1].Liq_Y) == 4
        for j in model.state[1].Liq_Y:
            if j in ["H+", "Na+"]:
                assert str(model.state[1].Liq_Y[j].expr) == str(
                    model.state[1].Liq_X[j]
                    / (model.state[1].Liq_X["Na+"] + model.state[1].Liq_X["H+"])
                )
            else:
                assert str(model.state[1].Liq_Y[j].expr) == str(
                    model.state[1].Liq_X[j]
                    / (model.state[1].Liq_X["Cl-"] + model.state[1].Liq_X["OH-"])
                )

        assert isinstance(model.state[1].Liq_ionic_strength, Expression)
        assert len(model.state[1].Liq_ionic_strength) == 1
        assert str(model.state[1].Liq_ionic_strength.expr) == str(
            0.5
            * (
                model.params.get_component("Cl-").config.charge ** 2
                * model.state[1].mole_frac_phase_comp_true["Liq", "Cl-"]
                + model.params.get_component("OH-").config.charge ** 2
                * model.state[1].mole_frac_phase_comp_true["Liq", "OH-"]
                + model.params.get_component("Na+").config.charge ** 2
                * model.state[1].mole_frac_phase_comp_true["Liq", "Na+"]
                + model.params.get_component("H+").config.charge ** 2
                * model.state[1].mole_frac_phase_comp_true["Liq", "H+"]
            )
        )

        assert isinstance(model.state[1].Liq_ionic_strength_ref, Expression)
        assert len(model.state[1].Liq_ionic_strength_ref) == 1
        assert str(model.state[1].Liq_ionic_strength_ref.expr) == str(
            0.5
            * (
                model.params.get_component("Cl-").config.charge ** 2
                * model.state[1].Liq_x_ref["Cl-"]
                + model.params.get_component("OH-").config.charge ** 2
                * model.state[1].Liq_x_ref["OH-"]
                + model.params.get_component("Na+").config.charge ** 2
                * model.state[1].Liq_x_ref["Na+"]
                + model.params.get_component("H+").config.charge ** 2
                * model.state[1].Liq_x_ref["H+"]
            )
        )

        assert isinstance(model.state[1].Liq_vol_mol_solvent, Expression)
        assert len(model.state[1].Liq_vol_mol_solvent) == 1
        assert str(model.state[1].Liq_vol_mol_solvent.expr) == "1/(42.0*mol/m**3)"

        assert isinstance(model.state[1].Liq_relative_permittivity_solvent, Expression)
        assert len(model.state[1].Liq_relative_permittivity_solvent) == 1
        assert str(model.state[1].Liq_relative_permittivity_solvent.expr) == (
            str(model.params.get_component("H2O").relative_permittivity_liq_comp)
        )

        assert isinstance(model.state[1].Liq_A_DH, Expression)
        assert len(model.state[1].Liq_A_DH) == 1
        assert_units_equivalent(model.state[1].Liq_A_DH, pyunits.dimensionless)
        assert str(model.state[1].Liq_A_DH.expr) == str(
            (1 / 3)
            * (
                2
                * Constants.pi
                * Constants.avogadro_number
                / model.state[1].Liq_vol_mol_solvent
            )
            ** 0.5
            * (
                Constants.elemental_charge**2
                / (
                    4
                    * Constants.pi
                    * model.state[1].Liq_relative_permittivity_solvent
                    * Constants.vacuum_electric_permittivity
                    * Constants.boltzmann_constant
                    * model.state[1].temperature
                )
            )
            ** (3 / 2)
        )

        assert isinstance(model.state[1].Liq_log_gamma_pdh, Expression)
        assert len(model.state[1].Liq_log_gamma_pdh) == 6
        for j in model.state[1].Liq_log_gamma_pdh:
            assert j in ["H2O", "C6H12", "Na+", "H+", "Cl-", "OH-"]
            if j in ["H2O", "C6H12"]:
                assert str(model.state[1].Liq_log_gamma_pdh[j].expr) == str(
                    (
                        2
                        * model.state[1].Liq_A_DH
                        * model.state[1].Liq_ionic_strength ** (3 / 2)
                        / (1 + 14.9 * model.state[1].Liq_ionic_strength ** (1 / 2))
                    )
                )

        assert isinstance(model.state[1].Liq_log_gamma_lc_I, Expression)
        assert len(model.state[1].Liq_log_gamma_lc_I) == 6
        for k in model.state[1].Liq_log_gamma_lc_I:
            assert k in ["H2O", "C6H12", "Na+", "H+", "Cl-", "OH-"]

        assert isinstance(model.state[1].Liq_log_gamma_lc_I0, Expression)
        assert len(model.state[1].Liq_log_gamma_lc_I0) == 4
        for k in model.state[1].Liq_log_gamma_lc_I0:
            assert k in ["Na+", "H+", "Cl-", "OH-"]
            assert str(model.state[1].Liq_log_gamma_lc_I0[k].expr) != str(
                model.state[1].Liq_log_gamma_lc_I[k].expr
            )

        assert isinstance(model.state[1].Liq_log_gamma_lc, Expression)
        assert len(model.state[1].Liq_log_gamma_lc) == 6
        for k in model.state[1].Liq_log_gamma_lc:
            assert k in ["H2O", "C6H12", "Na+", "H+", "Cl-", "OH-"]
            if k in ["H2O", "C6H12"]:
                assert str(model.state[1].Liq_log_gamma_lc[k].expr) == str(
                    model.state[1].Liq_log_gamma_lc_I[k]
                )
            else:
                assert str(model.state[1].Liq_log_gamma_lc[k].expr) == str(
                    model.state[1].Liq_log_gamma_lc_I[k]
                    - model.state[1].Liq_log_gamma_lc_I0[k]
                )

        assert isinstance(model.state[1].Liq_log_gamma, Expression)
        assert len(model.state[1].Liq_log_gamma) == 6
        for k, v in model.state[1].Liq_log_gamma.items():
            assert str(model.state[1].Liq_log_gamma[k].expr) == str(
                model.state[1].Liq_log_gamma_pdh[k] + model.state[1].Liq_log_gamma_lc[k]
            )

    @pytest.mark.unit
    def test_alpha(self, model):
        assert isinstance(model.state[1].Liq_alpha, Expression)
        assert len(model.state[1].Liq_alpha) == 28

        # Molecule-molecule interactions
        assert str(model.state[1].Liq_alpha["H2O", "H2O"].expr) == str(
            model.params.Liq.alpha["H2O", "H2O"]
        )
        assert str(model.state[1].Liq_alpha["H2O", "C6H12"].expr) == str(
            model.params.Liq.alpha["H2O", "C6H12"]
        )
        assert str(model.state[1].Liq_alpha["C6H12", "H2O"].expr) == str(
            model.params.Liq.alpha["H2O", "C6H12"]
        )
        assert str(model.state[1].Liq_alpha["C6H12", "C6H12"].expr) == str(
            model.params.Liq.alpha["C6H12", "C6H12"]
        )

        # Molecule-ion interactions
        assert str(model.state[1].Liq_alpha["H2O", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["H2O", "Na+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["H2O", "Na+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H2O", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["H2O", "H+, Cl-"]
                + model.state[1].Liq_Y["OH-"] * model.params.Liq.alpha["H2O", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Na+", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["H2O", "Na+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["H2O", "Na+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H+", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["H2O", "H+, Cl-"]
                + model.state[1].Liq_Y["OH-"] * model.params.Liq.alpha["H2O", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H2O", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * model.params.Liq.alpha["H2O", "Na+, Cl-"]
                + model.state[1].Liq_Y["H+"] * model.params.Liq.alpha["H2O", "H+, Cl-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H2O", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * model.params.Liq.alpha["H2O", "Na+, OH-"]
                + model.state[1].Liq_Y["H+"] * model.params.Liq.alpha["H2O", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Cl-", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * model.params.Liq.alpha["H2O", "Na+, Cl-"]
                + model.state[1].Liq_Y["H+"] * model.params.Liq.alpha["H2O", "H+, Cl-"]
            )
        )
        assert str(model.state[1].Liq_alpha["OH-", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * model.params.Liq.alpha["H2O", "Na+, OH-"]
                + model.state[1].Liq_Y["H+"] * model.params.Liq.alpha["H2O", "H+, OH-"]
            )
        )

        assert str(model.state[1].Liq_alpha["C6H12", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["C6H12", "Na+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["C6H12", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["C6H12", "H+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["C6H12", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Na+", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["C6H12", "Na+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H+", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * model.params.Liq.alpha["C6H12", "H+, Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["C6H12", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["C6H12", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["C6H12", "H+, Cl-"]
            )
        )
        assert str(model.state[1].Liq_alpha["C6H12", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["C6H12", "Na+, OH-"]
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["C6H12", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Cl-", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["C6H12", "H+, Cl-"]
            )
        )
        assert str(model.state[1].Liq_alpha["OH-", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["C6H12", "Na+, OH-"]
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["C6H12", "H+, OH-"]
            )
        )

        # Ion-ion interactions
        assert str(model.state[1].Liq_alpha["Na+", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * 0.2
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["Na+, Cl-", "H+, Cl-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Na+", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"] * 0.2
                + model.state[1].Liq_Y["H+"]
                * model.params.Liq.alpha["Na+, OH-", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["H+", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["Na+, Cl-", "H+, Cl-"]
                + model.state[1].Liq_Y["H+"] * 0.2
            )
        )
        assert str(model.state[1].Liq_alpha["H+", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * model.params.Liq.alpha["Na+, OH-", "H+, OH-"]
                + model.state[1].Liq_Y["H+"] * 0.2
            )
        )
        assert str(model.state[1].Liq_alpha["Cl-", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * 0.2
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["Na+, Cl-", "Na+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["Cl-", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"] * 0.2
                + model.state[1].Liq_Y["OH-"]
                * model.params.Liq.alpha["H+, Cl-", "H+, OH-"]
            )
        )
        assert str(model.state[1].Liq_alpha["OH-", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * model.params.Liq.alpha["Na+, Cl-", "Na+, OH-"]
                + model.state[1].Liq_Y["OH-"] * 0.2
            )
        )
        assert str(model.state[1].Liq_alpha["OH-", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * model.params.Liq.alpha["H+, Cl-", "H+, OH-"]
                + model.state[1].Liq_Y["OH-"] * 0.2
            )
        )

        # Like species interactions
        assert ("Na+", "Na+") not in model.state[1].Liq_alpha
        assert ("Na+", "H+") not in model.state[1].Liq_alpha
        assert ("H+", "Na+") not in model.state[1].Liq_alpha
        assert ("H+", "H+") not in model.state[1].Liq_alpha
        assert ("Cl-", "Cl-") not in model.state[1].Liq_alpha
        assert ("Cl-", "OH-") not in model.state[1].Liq_alpha
        assert ("OH-", "Cl-") not in model.state[1].Liq_alpha
        assert ("OH-", "OH-") not in model.state[1].Liq_alpha

    @pytest.mark.unit
    def test_G(self, model):
        assert isinstance(model.state[1].Liq_G, Expression)
        assert len(model.state[1].Liq_G) == 28

        # Molecule-molecule interactions
        assert str(model.state[1].Liq_G["H2O", "H2O"].expr) == str(1.0)
        assert str(model.state[1].Liq_G["H2O", "C6H12"].expr) == str(
            exp(
                -model.params.Liq.alpha["H2O", "C6H12"]
                * model.params.Liq.tau["H2O", "C6H12"]
            )
        )
        assert str(model.state[1].Liq_G["C6H12", "H2O"].expr) == str(
            exp(
                -model.params.Liq.alpha["H2O", "C6H12"]
                * model.params.Liq.tau["C6H12", "H2O"]
            )
        )
        assert str(model.state[1].Liq_G["C6H12", "C6H12"].expr) == str(1.0)

        # Molecule-ion interactions
        assert str(model.state[1].Liq_G["H2O", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, Cl-"]
                    * model.params.Liq.tau["H2O", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, OH-"]
                    * model.params.Liq.tau["H2O", "Na+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H2O", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, Cl-"]
                    * model.params.Liq.tau["H2O", "H+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, OH-"]
                    * model.params.Liq.tau["H2O", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Na+", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, Cl-"]
                    * model.params.Liq.tau["Na+, Cl-", "H2O"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "H2O"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H+", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, Cl-"]
                    * model.params.Liq.tau["H+, Cl-", "H2O"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "H2O"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H2O", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, Cl-"]
                    * model.params.Liq.tau["H2O", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, Cl-"]
                    * model.params.Liq.tau["H2O", "H+, Cl-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H2O", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, OH-"]
                    * model.params.Liq.tau["H2O", "Na+, OH-"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, OH-"]
                    * model.params.Liq.tau["H2O", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Cl-", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, Cl-"]
                    * model.params.Liq.tau["Na+, Cl-", "H2O"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, Cl-"]
                    * model.params.Liq.tau["H+, Cl-", "H2O"]
                )
            )
        )
        assert str(model.state[1].Liq_G["OH-", "H2O"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "H2O"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["H2O", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "H2O"]
                )
            )
        )

        assert str(model.state[1].Liq_G["C6H12", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                    * model.params.Liq.tau["C6H12", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, OH-"]
                    * model.params.Liq.tau["C6H12", "Na+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["C6H12", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, Cl-"]
                    * model.params.Liq.tau["C6H12", "H+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, OH-"]
                    * model.params.Liq.tau["C6H12", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Na+", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                    * model.params.Liq.tau["Na+, Cl-", "C6H12"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "C6H12"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H+", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, Cl-"]
                    * model.params.Liq.tau["H+, Cl-", "C6H12"]
                )
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "C6H12"]
                )
            )
        )
        assert str(model.state[1].Liq_G["C6H12", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                    * model.params.Liq.tau["C6H12", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, Cl-"]
                    * model.params.Liq.tau["C6H12", "H+, Cl-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["C6H12", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, OH-"]
                    * model.params.Liq.tau["C6H12", "Na+, OH-"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, OH-"]
                    * model.params.Liq.tau["C6H12", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Cl-", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, Cl-"]
                    * model.params.Liq.tau["Na+, Cl-", "C6H12"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, Cl-"]
                    * model.params.Liq.tau["H+, Cl-", "C6H12"]
                )
            )
        )
        assert str(model.state[1].Liq_G["OH-", "C6H12"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "C6H12"]
                )
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["C6H12", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "C6H12"]
                )
            )
        )

        # Ion-ion interactions
        assert str(model.state[1].Liq_G["Na+", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["Na+, Cl-", "H+, Cl-"]
                    * model.params.Liq.tau["Na+, Cl-", "H+, Cl-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Na+", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                + model.state[1].Liq_Y["H+"]
                * exp(
                    -model.params.Liq.alpha["Na+, OH-", "H+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["H+", "Cl-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["Na+, Cl-", "H+, Cl-"]
                    * model.params.Liq.tau["H+, Cl-", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["H+"]
            )
        )
        assert str(model.state[1].Liq_G["H+", "OH-"].expr) == str(
            (
                model.state[1].Liq_Y["Na+"]
                * exp(
                    -model.params.Liq.alpha["Na+, OH-", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "Na+, OH-"]
                )
                + model.state[1].Liq_Y["H+"]
            )
        )
        assert str(model.state[1].Liq_G["Cl-", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["Na+, Cl-", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, Cl-", "Na+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["Cl-", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                + model.state[1].Liq_Y["OH-"]
                * exp(
                    -model.params.Liq.alpha["H+, Cl-", "H+, OH-"]
                    * model.params.Liq.tau["H+, Cl-", "H+, OH-"]
                )
            )
        )
        assert str(model.state[1].Liq_G["OH-", "Na+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["Na+, Cl-", "Na+, OH-"]
                    * model.params.Liq.tau["Na+, OH-", "Na+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
            )
        )
        assert str(model.state[1].Liq_G["OH-", "H+"].expr) == str(
            (
                model.state[1].Liq_Y["Cl-"]
                * exp(
                    -model.params.Liq.alpha["H+, Cl-", "H+, OH-"]
                    * model.params.Liq.tau["H+, OH-", "H+, Cl-"]
                )
                + model.state[1].Liq_Y["OH-"]
            )
        )

        # Like species interactions
        assert ("Na+", "Na+") not in model.state[1].Liq_G
        assert ("Na+", "H+") not in model.state[1].Liq_G
        assert ("H+", "Na+") not in model.state[1].Liq_G
        assert ("H+", "H+") not in model.state[1].Liq_G
        assert ("Cl-", "Cl-") not in model.state[1].Liq_G
        assert ("Cl-", "OH-") not in model.state[1].Liq_G
        assert ("OH-", "Cl-") not in model.state[1].Liq_G
        assert ("OH-", "OH-") not in model.state[1].Liq_G

    @pytest.mark.unit
    def test_tau(self, model):
        assert isinstance(model.state[1].Liq_tau, Expression)
        assert len(model.state[1].Liq_tau) == 28

        # Molecule-molecule interactions
        assert str(model.state[1].Liq_tau["H2O", "H2O"].expr) == str(
            model.params.Liq.tau["H2O", "H2O"]
        )
        assert str(model.state[1].Liq_tau["H2O", "C6H12"].expr) == str(
            model.params.Liq.tau["H2O", "C6H12"]
        )
        assert str(model.state[1].Liq_tau["C6H12", "H2O"].expr) == str(
            model.params.Liq.tau["C6H12", "H2O"]
        )
        assert str(model.state[1].Liq_tau["C6H12", "C6H12"].expr) == str(
            model.params.Liq.tau["C6H12", "C6H12"]
        )

        for i, j in model.state[1].Liq_tau:
            if (i, j) not in [
                ("H2O", "H2O"),
                ("H2O", "C6H12"),
                ("C6H12", "H2O"),
                ("C6H12", "C6H12"),
            ]:
                assert str(model.state[1].Liq_tau[i, j].expr) == str(
                    -log(model.state[1].Liq_G[i, j]) / model.state[1].Liq_alpha[i, j]
                )

        # Like species interactions
        assert ("Na+", "Na+") not in model.state[1].Liq_tau
        assert ("Na+", "H+") not in model.state[1].Liq_tau
        assert ("H+", "Na+") not in model.state[1].Liq_tau
        assert ("H+", "H+") not in model.state[1].Liq_tau
        assert ("Cl-", "Cl-") not in model.state[1].Liq_tau
        assert ("Cl-", "OH-") not in model.state[1].Liq_tau
        assert ("OH-", "Cl-") not in model.state[1].Liq_tau
        assert ("OH-", "OH-") not in model.state[1].Liq_tau


class TestProperties(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.params = GenericParameterBlock(**configuration)

        m.state = m.params.build_state_block([1])

        # Need to set a value of T for checking expressions later
        m.state[1].temperature.set_value(300)
        # Need to set a value of log_act_phase_solvents for checking osmotic pressure calculation
        m.state[1].log_act_phase_solvents["Liq"].set_value(-1.789578)

        return m

    @pytest.mark.unit
    def test_pressure_osm_phase(self, model):
        model.state[1].vol_mol_phase = Var(
            model.params.phase_list,
            initialize=18e-6,
            units=pyunits.m**3 / pyunits.mol,
        )

        assert_units_equivalent(model.state[1].pressure_osm_phase["Liq"], pyunits.Pa)
        assert len(model.state[1].pressure_osm_phase) == 1
        assert pytest.approx(
            value(-Constants.gas_constant * 300 * log(0.1670306) / 18e-6), rel=1e-6
        ) == value(model.state[1].pressure_osm_phase["Liq"])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example of using the eNRTL model to model solutions of NaBr in mixed solvents.

Reference:

[1] Song, Y. and Chen, C.-C., Symmetric Electrolyte Nonrandom Two-Liquid Activity
Coefficient Model, Ind. Eng. Chem. Res., 2009, Vol. 48, pgs. 7788–7797

Author: Andrew Lee
"""
import pyomo.environ as pyo

from idaes.core import AqueousPhase, Solvent, Apparent, Anion, Cation
from idaes.models.properties.modular_properties.eos.enrtl import ENRTL
from idaes.models.properties.modular_properties.base.generic_property import (
    StateIndex,
)
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.pure.electrolyte import (
    relative_permittivity_constant,
)


def rho_H2O(b, *args, **kwargs):
    """Assume constant density of pure water"""
    return 1000 / 18e-3 * pyo.units.mol / pyo.units.m**3


def rho_MeOH(b, *args, **kwargs):
    """Assume constant density for ethanol"""
    return 792 / 32e-3 * pyo.units.mol / pyo.units.m**3


def rho_EtOH(b, *args, **kwargs):
    """Assume constant density for methanol"""
    return 789.45 / 46e-3 * pyo.units.mol / pyo.units.m**3


configuration = {
    "components": {
        "H2O": {
            "type": Solvent,
            "dens_mol_liq_comp": rho_H2O,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (18e-3, pyo.units.kg / pyo.units.mol),
                "relative_permittivity_liq_comp": 78.54,
            },
        },
        "MeOH": {
            "type": Solvent,
            "dens_mol_liq_comp": rho_MeOH,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (32e-3, pyo.units.kg / pyo.units.mol),
                "relative_permittivity_liq_comp": 32.6146,
            },
        },
        "EtOH": {
            "type": Solvent,
            "dens_mol_liq_comp": rho_EtOH,
            "relative_permittivity_liq_comp": relative_permittivity_constant,
            "parameter_data": {
                "mw": (46e-3, pyo.units.kg / pyo.units.mol),
                "relative_permittivity_liq_comp": 24.113,
            },
        },
        "NaBr": {"type": Apparent, "dissociation_species": {"Na+": 1, "Br-": 1}},
        "Na+": {"type": Cation, "charge": +1},
        "Br-": {"type": Anion, "charge": -1},
    },
    "phases": {"Liq": {"type": AqueousPhase, "equation_of_state": ENRTL}},
    "base_units": {
        "time": pyo.units.s,
        "length": pyo.units.m,
        "mass": pyo.units.kg,
        "amount": pyo.units.mol,
        "temperature": pyo.units.K,
    },
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "parameter_data": {
        "Liq_alpha": {
            ("H2O", "EtOH"): 0.3031,
            ("H2O", "MeOH"): 0.2994,
            ("MeOH", "EtOH"): 0.3356,
            ("H2O", "Na+, Br-"): 0.2,
            ("MeOH", "Na+, Br-"): 0.2,
            ("EtOH", "Na+, Br-"): 0.1,
        },
        "Liq_tau": {
            ("H2O", "MeOH"): 1.4265,
            ("MeOH", "H2O"): -0.42864,
            ("H2O", "EtOH"): 2.2485,
            ("EtOH", "H2O"): -0.18514,
            ("MeOH", "EtOH"): -0.04394,
            ("EtOH", "MeOH"): 0.02147,
            ("H2O", "Na+, Br-"): 9.527,
            ("Na+, Br-", "H2O"): -4.790,
            ("MeOH", "Na+, Br-"): 5.910,
            ("Na+, Br-", "MeOH"): -3.863,
            ("EtOH", "Na+, Br-"): 6.118,
            ("Na+, Br-", "EtOH"): -4.450,
        },
    },
}
